import re
import urllib.parse

from .networking import Request
from .utils import (
    _m3u8_attr,
    _parse_m3u8_resolution,
    approx_filesize_from_tbr,
    random_id,
)


class ExtractorBase:
    _EXTRACTOR_CLASSES = None
    _IE_NAME = None
    _VALID_URL = None

    def __init__(self, md=None):
        from .host_finder import HOST_FINDER

        self.md = md
        self.impersonate = md.impersonate
        self.to_screen = md.to_screen
        self.write_debug = md.write_debug
        self.raise_error = md.raise_error
        self.show_warning = md.show_warning
        self.logger = md.logger
        self.network_verbose = md.opts.print_traffic
        self.write = self.logger.to_stdout
        self.host_finder = HOST_FINDER(md, self)

    @classmethod
    def suitable(cls, url: str) -> bool:
        return bool(cls._VALID_URL and re.match(cls._VALID_URL, url))

    def extract(self, url: str) -> dict:
        return self._real_extract(url)

    def _real_extract(self, url: str) -> dict:
        raise NotImplementedError

    @classmethod
    def import_extractors(cls):
        if cls._EXTRACTOR_CLASSES is not None:
            return
        from .extractors import extractors as ex

        cls._EXTRACTOR_CLASSES = ex.ALL

    @classmethod
    def gen_extractors(cls, md):
        cls.import_extractors()
        return [ie_cls(md) for ie_cls in cls._EXTRACTOR_CLASSES]

    @classmethod
    def get_extractor(cls, md, url: str):
        for ie in cls.gen_extractors(md):
            if ie.suitable(url):
                return ie
        return None

    def _request(
        self,
        url,
        *,
        method="GET",
        headers=None,
        cookies=None,
        query=None,
        data=None,
        timeout=20,
        note=None,
    ):
        if note:
            self.to_screen(note)

        kw = dict(
            url=url,
            method=method,
            headers=headers,
            cookies=cookies,
            query=query,
            data=data,
            timeout=timeout,
            impersonate=self.impersonate,
        )

        return Request(**kw, verbose=self.network_verbose, logger=self.logger)()

    def _download_webpage(self, url, note="Downloading Webpage", **kw):
        return self._request(url, method="GET", **kw, note=note).text

    def _download_json(self, url, note="Downloading Json", **kw):
        method = "POST" if (kw.get("data") or {}) else "GET"
        return self._request(url, method=method, **kw, note=note).json()

    def _ping_host(self, url, note="ping Host", **kw):
        return self._request(url, method="GET", **kw, note=note).status_code

    def extract_m3u8_formats(self, m3u8_url, label="hls", headers=None, duration=None):
        m3u8 = self._download_webpage(
            m3u8_url, headers=headers or {}, note="Downloading m3u8 information"
        )
        if not m3u8 or not m3u8.lstrip().startswith("#EXTM3U"):
            return []

        if "#EXT-X-STREAM-INF" not in m3u8:
            fmt = [
                {
                    "url": m3u8_url,
                    "label": random_id(),
                    "ext": "mp4",
                    "vcodec": True,
                    "acodec": True,
                }
            ]
            if headers:
                fmt["http_headers"] = headers

            return fmt

        out = []
        last = None
        for line in m3u8.splitlines():
            line = line.strip()
            if line.startswith("#EXT-X-STREAM-INF:"):
                last = line
                continue
            if not line or line.startswith("#") or not last:
                continue

            u = urllib.parse.urljoin(m3u8_url, line)
            w, h = _parse_m3u8_resolution(last)
            bw = _m3u8_attr(last, "BANDWIDTH")
            tbr = int(bw) // 1000 if (bw and bw.isdigit()) else None

            fmt = {
                "url": u,
                "format_id": (f"{h or w or 'var'}"),
                "width": w,
                "height": h,
                "tbr": tbr,
                "ext": "mp4",
                "vcodec": True,
                "acodec": True,
                "format_note": label,
            }
            if headers:
                fmt["http_headers"] = headers
            if duration:
                approx = approx_filesize_from_tbr(tbr, duration)
                if approx:
                    fmt["filesize_approx"] = approx
            out.append(fmt)
            last = None

        return out

    def playlist_result(
        self,
        entries,
        playlist_url=None,
        playlist_title=None,
        playlist_description=None,
        **kwargs,
    ):
        if playlist_url:
            kwargs["url"] = playlist_url
        if playlist_title:
            kwargs["title"] = playlist_title
        if playlist_description is not None:
            kwargs["description"] = playlist_description
        return {
            **kwargs,
            "_type": "series",
            "entries": entries,
        }



import sys

from .utils import NobiDLError


class Log:
    def __init__(self, params=None, out=None, err=None):
        self.params = params
        self._out = out or sys.stdout
        self._err = err or sys.stderr
        self._printed_once = set()
        self._ctx_ie = None

    def _param(self, key, default=None):
        p = self.params
        if p is None:
            return default
        if isinstance(p, dict):
            return p.get(key, default)
        return getattr(p, key, default)

    def _write(self, s, stream, only_once=False):
        if only_once:
            if s in self._printed_once:
                return
            self._printed_once.add(s)
        stream.write(s)
        stream.flush()

    def to_stdout(self, s: str, *, only_once=False, newline=True):
        if s is None:
            return
        if not isinstance(s, str):
            s = str(s)
        if newline and not s.endswith("\n"):
            s += "\n"
        self._write(s, stream=self._out, only_once=only_once)

    def to_stderr(self, s: str, *, only_once=False, newline=True):
        if s is None:
            return
        if newline and not s.endswith("\n"):
            s += "\n"
        self._write(s, stream=self._err, only_once=only_once)

    def set_ie(self, ie, suffix=None):
        self._ctx_ie = ie
        self.suffix = suffix

    def _ie(self) -> str:
        ie = self._ctx_ie
        suffix = self.suffix
        if not ie:
            return f"[{suffix}] " if suffix else ""
        name = getattr(ie, "_IE_NAME", None) or ie.__class__.__name__
        if suffix is not None:
            return f"[{name}][{self.suffix}] "
        return f"[{name}] "

    def to_screen(self, message, *, only_once=False):
        if message is None:
            return
        if self._param("quiet", False) and not self._param("verbose", 0):
            return
        ie = self._ie()
        if not ie:
            ie = ""
        self.to_stdout(f"{ie}{message}", only_once=only_once)

    def write_debug(self, message):
        if message is None:
            return
        if self._param("verbose", 0):
            self.to_stderr(f"[debug] {message}")

    def warning(self, message: str):
        prefix = "[WARNING]"
        if getattr(self._err, "isatty", lambda: False)():
            prefix = "\033[33m[WARNING]\033[0m"
        self.to_stderr(f"{prefix}: {message}", only_once=True)

    def error(self, message: str, fatal=False):
        if fatal is True:
            self.warning(message=message)
        raise NobiDLError(msg=message)



import http.client
import json
import socket
import urllib.error
import urllib.parse
import urllib.request


class HttpResponse:
    def __init__(self, resp, body: bytes):
        self._resp = resp
        self._body = body
        self._headers = dict(resp.headers)

    @property
    def content(self):
        return self._body

    @property
    def text(self):
        return self._body.decode("utf-8", errors="ignore")

    def json(self):
        return json.loads(self.text)

    def iter_content(self, chunk_size=8192):
        if self._body:
            for i in range(0, len(self._body), chunk_size):
                yield self._body[i : i + chunk_size]
            return

        while True:
            chunk = self._resp.read(chunk_size)
            if not chunk:
                break
            yield chunk

    @property
    def url(self):
        return getattr(self._resp, "url", None)

    @property
    def status_code(self):
        return getattr(self._resp, "status", None) or self._resp.getcode()

    @property
    def headers(self):
        return self._headers


class UrllibRH:
    def __init__(
        self,
        url,
        headers=None,
        cookies=None,
        query=None,
        data=None,
        timeout=20,
        method="GET",
        retries=2,
        verbose=False,
        logger=None,
        stream=False,
    ):
        self.logger = logger
        self.verbose = verbose
        self.url = url
        self.headers = headers or {}
        self.cookies = cookies or {}
        self.query = query or {}
        self.data = data
        self.timeout = timeout or 20
        self.method = (method or "GET").upper()
        self.retries = retries
        self.stream = stream

    def __call__(self):
        return self._request()

    def _print_verbose(self, msg):
        if self.verbose and self.logger:
            self.logger.to_stdout(msg)

    def print_traffic(self):
        if not self.verbose:
            return
        self._print_verbose("[info] Sending Request via urllib")
        http.client.HTTPSConnection.debuglevel = 1
        http.client.HTTPConnection.debuglevel = 1

    def _build_url(self) -> str:
        if not self.query:
            return self.url
        parts = list(urllib.parse.urlsplit(self.url))
        q = dict(urllib.parse.parse_qsl(parts[3], keep_blank_values=True))
        q.update(self.query)
        parts[3] = urllib.parse.urlencode(q, doseq=True)
        return urllib.parse.urlunsplit(parts)

    @property
    def update_cookies_in_headers(self):
        if not self.cookies:
            return {}

        if isinstance(self.cookies, dict):
            cookie_header = "; ".join(f"{k}={v}" for k, v in self.cookies.items())
        else:
            cookie_header = str(self.cookies)

        return {"Cookie": cookie_header}

    def _request(self):
        from ._request_handler import std_headers

        self.print_traffic()

        default_headers = dict(std_headers)
        default_headers.update(self.headers)
        default_headers.update(self.update_cookies_in_headers)

        url = self._build_url()

        body = None
        if self.data is not None:
            if isinstance(self.data, (bytes, bytearray)):
                body = bytes(self.data)
            elif isinstance(self.data, str):
                body = self.data.encode("utf-8")
            elif isinstance(self.data, dict):
                body = urllib.parse.urlencode(self.data, doseq=True).encode("utf-8")
            else:
                body = str(self.data).encode("utf-8")

        last_err = None
        for _ in range(self.retries + 1):
            try:
                req = urllib.request.Request(
                    url=url,
                    data=body,
                    headers=default_headers,
                    method=self.method,
                )
                resp = urllib.request.urlopen(req, timeout=self.timeout)
                code = getattr(resp, "status", None) or resp.getcode()

                if code in (429, 503, 502, 504):
                    last_err = RuntimeError(f"HTTP {code}")
                    continue

                if self.stream:
                    return HttpResponse(resp, b"")

                data = resp.read()
                resp.close()
                return HttpResponse(resp, data)

            except urllib.error.HTTPError as e:
                if e.code in (429, 503, 502, 504):
                    last_err = RuntimeError(f"HTTP {e.code}")
                    continue
                raise
            except (urllib.error.URLError, socket.timeout) as e:
                last_err = e
                continue

        raise last_err



import json
import tempfile
import time
from pathlib import Path

from .utils import _search_regex, b64d, clean_url

CACHE_FILE = Path(tempfile.gettempdir()) / "movie_dl_cache.json"
TTL_SECONDS = 24 * 60 * 60


class HOST_FINDER:
    def __init__(self, md, extractor):
        self.md = md
        self.extractor = extractor
        self.impersonate = md.impersonate
        self.to_screen = md.to_screen
        self.write_debug = md.write_debug
        self.raise_error = md.raise_error
        self.show_warning = md.show_warning
        self.logger = md.logger
        self.network_verbose = md.opts.print_traffic
        self.write = self.logger.to_stdout
        self._request = extractor._request
        self._download_webpage = extractor._download_webpage
        self._download_json = extractor._download_json

    DEFAULT_HOSTS = {
        "4khub": "https://4khdhub.dad",
        "hdhub": "https://new3.hdhub4u.fo",
        "bollyflix": "https://bollyflix.sarl",
        "moviesmod": "https://moviesmod.town",
        "vega": "https://vegamovies13.com",
    }

    def _read_cache(self) -> dict:
        if not CACHE_FILE.exists():
            return {}
        try:
            return json.loads(CACHE_FILE.read_text(encoding="utf-8")) or {}
        except Exception:
            return {}

    def _write_cache(self, data: dict):
        CACHE_FILE.write_text(json.dumps(data, indent=2), encoding="utf-8")

    def _get_cached_host(self, name: str, no_ttl_check=False) -> str | None:
        cache = self._read_cache()
        item = cache.get(name) or {}
        host = item.get("host")
        if no_ttl_check is True:
            return host
        exp = item.get("expires_at", 0)
        if isinstance(host, str) and host and exp > time.time():
            return host
        return None

    def _set_cached_host(self, name: str, host: str):
        cache = self._read_cache()
        cache[name] = {"host": host, "expires_at": time.time() + TTL_SECONDS}
        self._write_cache(cache)

    def _resolve_and_cache(self, name: str, resolver):
        host = self._get_cached_host(name)
        if host:
            return host
        host = resolver()
        if isinstance(host, str) and host:
            self._set_cached_host(name, host)
            return host
        return None

    def _resolve_vega(self) -> str | None:
        old_domain = self._get_cached_host("vega", True)
        if not old_domain:
            old_domain = self.DEFAULT_HOSTS.get("vega")
        try:
            req = self._request(old_domain)
            code = getattr(req, "status_code", None)
            final_url = getattr(req, "url", None)
            if code in (301, 302) and final_url:
                return clean_url(final_url)
        except Exception:
            return self.DEFAULT_HOSTS.get("vega")
        return old_domain

    def _resolve_hdhub(self) -> str | None:
        data = self._download_json(
            "https://cdn.hub4u.cloud/host/?v=2026011617",
            headers={"referer": "https://hdhub4u.gd/", "origin": "https://hdhub4u.gd"},
        )
        c = (data or {}).get("c")
        return clean_url(b64d(c)) if c else None

    def _resolve_4khub(self) -> str | None:
        hh = self.hdhub_host()
        if not hh:
            return None
        webpage = self._download_webpage(hh)
        return _search_regex(r'((?:https://)?4khd[^"]+)', webpage)

    def _resolve_bollyflix(self) -> str | None:
        old_domain = self._get_cached_host("bollyflix", True)
        if not old_domain:
            old_domain = self.DEFAULT_HOSTS.get("bollyflix")
        try:
            req = self._request(old_domain)
            code = getattr(req, "status_code", None)
            final_url = getattr(req, "url", None)
            if code in (301, 302, 303, 307, 308) and final_url:
                return clean_url(final_url)
        except Exception:
            return self.DEFAULT_HOSTS.get("bollyflix")
        return old_domain

    def _resolve_moviemod(self) -> str | None:
        old_domain = self._get_cached_host("moviesmod", True)
        if not old_domain:
            old_domain = self.DEFAULT_HOSTS.get("moviesmod")
        try:
            req = self._request(old_domain)
            code = getattr(req, "status_code", None)
            final_url = getattr(req, "url", None)
            if code in (301, 302, 303, 307, 308) and final_url:
                return clean_url(final_url)
        except Exception:
            return self.DEFAULT_HOSTS.get("moviesmod")
        return old_domain

    def vega_host(self) -> str | None:
        return self._resolve_and_cache("vega", self._resolve_vega)

    def hdhub_host(self) -> str | None:
        return self._resolve_and_cache("hdhub", self._resolve_hdhub)

    def _4k_host(self) -> str | None:
        return self._resolve_and_cache("4khub", self._resolve_4khub)

    def bollyflix(self) -> str | None:
        return self._resolve_and_cache("bollyflix", self._resolve_bollyflix)

    def moviemod(self) -> str | None:
        return self._resolve_and_cache("moviesmod", self._resolve_moviemod)

    def _hdhub_search_host(self) -> str | None:
        host = self.hdhub_host()
        if not host:
            return None
        webpage = self._download_webpage(f"{host}/search.html")
        search_host = _search_regex(r'https?://(search[^/]+)/[^",]+', webpage)
        if not search_host:
            return None
        return clean_url(search_host)



import json
import os
import time

from ..networking import Request
from ..utils import filename_from_title
from .common import ProgressPrinter


class HlsDL:
    def __init__(self, format, info_dict, options, logger):
        self.info_dict = info_dict
        self.format = format
        self.opts = options
        self.logger = logger
        self.to_stdout = logger.to_stdout
        self.continue_dl = self.opts.continue_dl
        self.printer = ProgressPrinter(self.logger)

    def __call__(self):
        return self.downloader_hls()

    def to_screen(self, msg):
        return self.to_stdout(f"[hls_native] {msg}")

    def extract_all_hls_fragments(self, text: str) -> list[str]:
        fragments = []
        lines = text.splitlines()
        expect_url = False

        for line in lines:
            line = line.strip()
            if not line:
                continue

            if line.startswith("#EXTINF"):
                expect_url = True
                continue

            if expect_url and not line.startswith("#"):
                fragments.append(line)
                expect_url = False

        return fragments

    def hls_data(self, url, headers=None):
        headers = headers or {}
        self.to_screen("Downloading m3u8 information")
        hls_data = Request(url, headers=headers)()
        hls_data = hls_data.text
        return hls_data

    def nobidl_filename(self, filename):
        filename = filename.replace(".part", "") if ".part" in filename else filename
        return filename + ".nobidl"

    def read_nobidl_data(self, filename):
        if not os.path.exists(filename):
            return {}

        try:
            with open(filename, "r") as f:
                return json.load(f)
        except Exception:
            return {}

    def max_existing_fragment(self, filename):
        base = filename + ".part-Frag"
        found = []

        for f in os.listdir("."):
            if f.startswith(os.path.basename(base)) and f.endswith(".part"):
                try:
                    idx = int(f.split("Frag")[1].split(".")[0])
                    found.append(idx)
                except Exception:
                    pass

        return max(found) if found else None

    def check_continue_dl_and_files(self, filename):
        if not self.continue_dl:
            return 0

        nobidl = self.nobidl_filename(filename)
        nobi_data = self.read_nobidl_data(nobidl)
        disk_idx = self.max_existing_fragment(filename)
        if not nobi_data:
            return (disk_idx + 1) if disk_idx is not None else 0
        json_idx = nobi_data.get("current_fragment", 0)
        if disk_idx is None:
            return json_idx

        return min(json_idx, disk_idx + 1)

    def write_nobi_data(self, filename, curr_frag):
        nobi_filename = (
            self.nobidl_filename(filename)
            if not filename.endswith(".nobidl")
            else filename
        )

        tmp = nobi_filename + ".tmp"
        with open(tmp, "w") as wr:
            json.dump(
                {
                    "note": "Do not delete this file if you want to resume your download",
                    "current_fragment": curr_frag,
                    "written_by": "NOBI_DL",
                },
                wr,
                indent=2,
            )

        os.replace(tmp, nobi_filename)
        return True

    def frag_filename(self, index: int) -> str:
        return f"Frag{index:06d}.part"

    def remove_already_downloaded_fragments(
        self,
        total_frag: list,
        last_idx_frag: int | None,
    ) -> list:
        if last_idx_frag is None:
            return total_frag
        return total_frag[last_idx_frag + 1 :]

    def write_chunk_in_file(self, resp, filename):
        with open(filename, "ab") as f:
            it = getattr(resp, "iter_bytes", None)
            if it:
                for chunk in it(1024 * 64):
                    if chunk:
                        f.write(chunk)
            else:
                for chunk in resp.iter_content(1024 * 64):
                    if chunk:
                        f.write(chunk)

    def remove_frag_file(self, filename):
        os.remove(filename)

    def downloader_hls(self):
        try:
            bfmt = self.format
            title = self.info_dict.get("title")
            url = bfmt.get("url")
            ext = "mp4"
            format_id = bfmt.get("format_id")

            filename = (
                filename_from_title(title, url, ext)
                if title
                else f"Unknown_movie.{ext}"
            )
            part = filename + ".hls.part"

            self.to_screen(f"Downloading format: {format_id}")
            self.to_screen(f'Invoking Downloader on "{url}"')
            self.to_screen(f"Destination: {filename}")

            headers = bfmt.get("http_headers") or {}

            hls_data = self.hls_data(url, headers=headers)
            all_fragments = self.extract_all_hls_fragments(hls_data)

            total_frags = len(all_fragments)
            self.to_screen(f"Total Fragments: {total_frags}")

            resume_idx = self.check_continue_dl_and_files(filename)
            if resume_idx:
                self.to_screen(f"Resume at {resume_idx} fragment")
            fragments = self.remove_already_downloaded_fragments(
                all_fragments, resume_idx
            )

            downloaded = resume_idx

            last_t = time.time()
            last_done = resume_idx
            ema_speed = None
            alpha = 0.2
            for idx, frag_url in enumerate(fragments, start=resume_idx):
                frag_file = self.frag_filename(idx)

                resp = Request(frag_url, headers=headers, stream=True)()
                self.write_chunk_in_file(resp, frag_file)

                with open(part, "ab") as out, open(frag_file, "rb") as frag:
                    out.write(frag.read())

                self.remove_frag_file(frag_file)
                self.write_nobi_data(filename, idx)

                downloaded += 1
                now = time.time()
                dt = now - last_t
                df = (idx + 1) - last_done

                if dt > 0:
                    inst_speed = df / dt
                    if ema_speed is None:
                        ema_speed = inst_speed
                    else:
                        ema_speed = alpha * inst_speed + (1 - alpha) * ema_speed
                else:
                    inst_speed = None

                eta = (
                    ((total_frags - (idx + 1)) / ema_speed)
                    if ema_speed and ema_speed > 0
                    else None
                )

                self.printer(
                    {
                        "status": "downloading",
                        "progress_type": "fragments",
                        "downloaded_fragments": idx + 1,
                        "total_fragments": total_frags,
                        "speed": ema_speed,
                        "eta": eta,
                    }
                )

                last_t = now
                last_done = idx + 1

            os.replace(part, filename)
            os.remove(self.nobidl_filename(filename))
            self.printer(
                {
                    "status": "finished",
                    "progress_type": "fragments",
                    "downloaded_fragments": total_frags,
                    "total_fragments": total_frags,
                    "speed": None,
                    "eta": None,
                }
            )

        except Exception:
            raise SystemExit(130)

        try:
            if resp and getattr(resp, "close", None):
                resp.close()
        except Exception:
            pass

        except KeyboardInterrupt:
            if self.printer:
                self.printer(
                    {
                        "status": "error",
                        "progress_type": "fragments",
                        "downloaded_fragments": total_frags,
                        "total_fragments": total_frags,
                        "speed": None,
                        "eta": None,
                    }
                )
            self.to_screen(
                "Interrupted by user. add -c or --continue for continue downloading."
            )
            raise SystemExit(130)



import itertools
import re

from nobi_dl import ExtractorBase
from nobi_dl.utils import (
    _og_thumbnail,
    _og_title,
    _parse_a_tag,
    _parse_a_tags,
    _parse_resolution,
    _search_json,
    _search_regex,
    determine_ext,
    determine_filesize,
    fix_entries,
    is_series,
    random_id,
)


class VegaMoviesME(ExtractorBase):
    _IE_NAME = "vegamovies"
    _VALID_URL = r"https://?[^\.]+?\.?vegamovies[^/]+/[^?&]+"
    _SEARCH = True

    @property
    def HOST(self):
        return self.host_finder.vega_host()

    def _parse_content(self, html):
        results = []
        for article in re.findall(r"<article[\s\S]+?<\/article>", html):
            href = _search_regex(r'href="([^"]+)"', article)
            href = f"{self.HOST}{href}" if "https" not in href else href
            thumbnail = _search_regex(r'src="([^"]+)"', article)
            thumbnail = (
                f"{self.HOST}{thumbnail}" if "https" not in thumbnail else thumbnail
            )
            title = _search_regex(r'title="([^"]+)"', article)
            results.append(
                {
                    "title": title,
                    "thumbnail": thumbnail,
                    "url": href,
                    "quality": _parse_resolution(title),
                    "source": "vegamovies",
                }
            )
        return results

    def _real_search(self, query, limit=50):
        results = []

        for pagenum in itertools.count(1):
            page = self._download_webpage(
                self.HOST,
                query={
                    "do": "search",
                    "subaction": "search",
                    "story": query,
                    "search_start": str(pagenum),
                },
                headers={
                    "referer": f"{self.HOST}/",
                },
            )
            results.extend(self._parse_content(page))
            next_page = _search_regex(r"<a[^>]+list_submit[^>]+>(\d+)<\/a>", page)
            if not next_page or len(results) >= limit:
                break

        return results

    def _parse_jekop_quailty(self, format, domain, headers, duration=None):
        file = format.get("file")
        title = format.get("title")
        if not file:
            return []
        file = file[1:] if file.startswith("~") else file
        file_url = f"{domain}/playlist/{file}.txt"
        file_data = self._download_webpage(file_url, headers=headers)
        if ".m3u8" not in file_data:
            return []
        return self.extract_m3u8_formats(
            file_data, title, headers=headers, duration=duration
        )

    def _extract_jekop_formats(self, webpage, duration=None):
        formats = []
        is_video_player = _search_regex(
            r"(?:playerconfigs|PlayerConfigs)\s*=\s*\{([\s\S]+?)}", webpage
        )
        if not is_video_player:
            self.to_screen("No Player for this movie")
            return []

        sid = _search_regex(r"""\bsrc\s*:\s*['\"]([^'\"]+)['\"]""", webpage)
        if not sid:
            self.to_screen("No Player formats for this movie")
            return []

        player_url = _search_regex(
            r"(https?:)?//[^\"'\s]+/player[^\"'\s]+", webpage, default=None
        )
        if not player_url:
            self.show_warning("Unable to find jekop player url")
            return []

        if player_url.startswith("//"):
            player_url = "https:" + player_url
        elif not player_url.startswith("http"):
            player_url = "https://" + player_url
        player_data = self._download_webpage(
            player_url, note="Downloading Jekop JS Player"
        )
        domain = _search_regex(r"""(https://?[^"']+)""", player_data)
        domain = f"https://{domain}" if "http" not in domain else domain
        video_url = f"{domain}/play/{sid}"
        video_webpage = self._download_webpage(
            video_url,
            headers={
                "Referer": "https://vegavinc.com/",
            },
            note="Downloading Jekop video webpage",
        )
        video_data = _search_json(r"=\s*(\{[\s\S]+?);", video_webpage)
        file_url = video_data.get("file")
        if not file_url:
            self.show_warning("Unable to Solve jekop data")
            return []
        common_headers = {
            "Referer": "https://vegavinc.com/",
            "X-CSRF-TOKEN": video_data.get("key"),
        }
        file_data = self._download_json(
            file_url,
            headers=common_headers,
            note="Downloading Multi Video Jekop information",
        )
        for file in file_data:
            formats.extend(
                self._parse_jekop_quailty(
                    file, domain, common_headers, duration=duration
                )
            )
        return formats

    def _get_movie(self, url_or_webpage):
        formats = []

        if isinstance(url_or_webpage, str) and "<" in url_or_webpage:
            webpage = url_or_webpage
        else:
            webpage = self._download_webpage(url_or_webpage)

        for h3 in re.findall(r"(<h3[\s\S]+?<\/h3>\s*(?:<h3[\s\S]+?<\/h3>)?)", webpage):
            if not any(k in h3 for k in ("nexdrive", "fast-dl", "st-dl", "extdrive")):
                continue
            h3_text = _search_regex(r"<h3[^>]+>(?:<[^>]+>)([^<]+)<", h3)
            for atag in _parse_a_tags(h3):
                href, name = _parse_a_tag(atag)
                if not href:
                    continue

                label = (name or "").lower()

                if not any(
                    k in label
                    for k in ("download", "to download", "click here to download")
                ):
                    continue

                formats.append(
                    {
                        "format_id": random_id(),
                        "url": href,
                        "acodec": True,
                        "vcodec": True,
                        "ext": determine_ext(href, href),
                        **determine_filesize(h3),
                        **_parse_resolution(h3_text),
                    }
                )

        if not formats:
            formats.extend(self._extract_jekop_formats(webpage))

        return {
            "title": _og_title(webpage),
            "thumbnail": _og_thumbnail(webpage),
            "formats": formats,
        }

    def extract_series_formats(self, url, op_webpage):
        webpage = self._download_webpage(url)
        title = _og_title(op_webpage)
        blocks = []
        for h4 in re.findall(r"<h4[\s\S]+?<\/h4>\s*<div[\s\S]+?<\/div>", webpage):
            h4_title = _search_regex(
                r"<h4[^>]+>\s*(?:<[^>]+>)?(?:<[^>]+>)?([^<]+)<", h4
            )
            h4_title = re.sub(r"(?:-:|:-)", "", h4_title)
            m = re.search(r"episodes?\s*[:\-]?\s*(\d+)", h4_title, re.I)
            ep_no = int(m.group(1))
            ep_key = f"EP {ep_no:02d}"
            season = _search_regex(r"(?is)(?:season|s0|s)\s*(\d+)", title, default=1)
            season
            for atag in _parse_a_tags(h4):
                href, name = _parse_a_tag(atag)
                label = (name or "").lower()
                if not any(k in label for k in ("g-direct", "direct")):
                    continue
                blocks.append(
                    {
                        "title": f"{title} - season {season} - {ep_key}",
                        "thumbanail": _og_thumbnail(op_webpage),
                        "episode": ep_no,
                        "season": season,
                        "formats": [
                            {
                                "format_id": random_id(),
                                "format_note": h4_title,
                                "url": href,
                                "acodec": True,
                                "vcodec": True,
                                "ext": "mkv",
                                **determine_filesize(name),
                                **_parse_resolution(webpage),
                            }
                        ],
                    }
                )

        return blocks

    def _get_series(self, url_or_name_or_webpage):
        if "<html" in url_or_name_or_webpage:
            webpage = url_or_name_or_webpage
        else:
            webpage = self._download_webpage(url_or_name_or_webpage)

        def entries(webpage):
            entries = []
            for atag in _parse_a_tags(webpage):
                href, name = _parse_a_tag(atag)
                format_id = (name or "").lower()
                if not any(
                    k in format_id
                    for k in (
                        "click here to download",
                        "to download",
                        "here to download",
                    )
                ):
                    continue
                entries.extend(self.extract_series_formats(href, webpage))
            return entries

        new_entires = fix_entries(self, entries(webpage))

        return self.playlist_result(
            entries=new_entires, playlist_title=_og_title(webpage)
        )

    def _real_extract(self, url):
        webpage = self._download_webpage(url)
        if is_series(url, webpage):
            return self._get_series(webpage)
        else:
            return self._get_movie(webpage)



import base64
import difflib
import json
import os
import random
import re
import sys
import urllib.parse
from html import unescape

from .version import REPOSITORY

try:
    from rapidfuzz import fuzz
except ImportError:
    fuzz = None


DEFAULT = None

KNOWN_RESOLUTIONS = {
    256: 144,
    426: 240,
    640: 360,
    854: 480,
    960: 540,
    1280: 720,
    1600: 900,
    1920: 1080,
    2560: 1440,
    3840: 2160,
    144: 256,
    240: 426,
    360: 640,
    480: 854,
    540: 960,
    720: 1280,
    900: 1600,
    1080: 1920,
    1440: 2560,
    2160: 3840,
    0: 0,
    None: 0,
}

JUNK_C = re.compile(
    r"(?i)\b("
    r"blu ?ray|brrip|hdrip|webrip|web[- ]dl|hdtv|x264|x265|hevc|aac|ddp?5\.?1|atmos|"
    r"org|nf|proper|repack|uncut|full movie|movie|"
    r"hindi|english|bengali|tamil|telugu|dual audio|"
    r"mp4|mkv|"
    r"\d{3,4}p|\d{3,4}x\d{3,4}|4k|8k|"
    r"\d+(?:\.\d+)?\s?(?:gb|mb|mib)|\d+k|"
    r"note|use|download|manager|instant|direct|"
    r"idm|adm|g\s?direct|like|or|for|g|g-direct"
    r")\b"
)

YEAR_C = re.compile(r"\b(19\d{2}|20\d{2})\b")

SERIES_URL_PAT = re.compile(
    r"/(season-\d+|s\d+e\d+|episode-\d+|ep-?\d+|all-episodes)\b", re.I
)
STRONG_SERIES = re.compile(
    r"\b(s\d+\s*e\d+|season\s*\d+|episode\s*\d+|all\s+episodes|web\s*series)\b", re.I
)
MOVIE_HINTS = re.compile(r"\b(full\s*movie|bluray|brrip|hdrip|webrip|dvdrip)\b", re.I)

_UNITS = {
    "b": 1,
    "kb": 10**3,
    "mb": 10**6,
    "gb": 10**9,
    "tb": 10**12,
    "pb": 10**15,
    "kib": 2**10,
    "mib": 2**20,
    "gib": 2**30,
    "tib": 2**40,
    "pib": 2**50,
}

SIZE_RE = re.compile(
    r"(?i)\b(?:size|filesize)\s*[:=]\s*([\d.,]+)\s*(b|kb|mb|gb|tb|pb|kib|mib|gib|tib|pib)\b"
)
INLINE_SIZE_RE = re.compile(
    r"(?i)(?<![a-z0-9.-])\b([\d.]{1,6})\s*(gb|mb|kb|tb|pb|gib|mib|kib|tib|pib)\b"
)
BADGE_SIZE_RE = re.compile(
    r'(?i)<span[^>]*class="[^"]*badge-size[^"]*"[^>]*>\s*([\d.,]+)\s*(b|kb|mb|gb|tb|pb|kib|mib|gib|tib|pib)\s*</span>'
)
CONTENT_LEN_RE = re.compile(r"(?i)\bcontent-length\s*[:=]\s*(\d+)\b")

KNOWN_EXTENSIONS = {
    "mp4",
    "mkv",
    "mov",
    "webm",
    "flv",
    "avi",
    "m4v",
    "ogv",
    "mpg",
    "mpeg",
    "m3u8",
    "ts",
}


def _parse_resolution(webpage_or_url: str):
    q = _search_regex(r"\b(\d{3,4})p\b", webpage_or_url, default=None)
    if q:
        h = int(q)
        w = KNOWN_RESOLUTIONS.get(h)
        return {"height": h, "width": w} if w else {}

    q = _search_regex(
        r"(?is)<title[^>]*>.*?\b(\d{3,4})p\b", webpage_or_url, default=None
    )
    if q:
        h = int(q)
        w = KNOWN_RESOLUTIONS.get(h)
        return {"height": h, "width": w} if w else {}

    q = re.search(r"\b(\d{3,4})\s*[xX]\s*(\d{3,4})\b", webpage_or_url or "")
    if not q:
        return {}

    a, b = int(q.group(1)), int(q.group(2))
    if a in KNOWN_RESOLUTIONS and KNOWN_RESOLUTIONS[a] == b:
        return {"height": a, "width": b}
    if b in KNOWN_RESOLUTIONS and KNOWN_RESOLUTIONS[b] == a:
        return {"height": b, "width": a}
    return {}


def b64d(s: str) -> str:
    s = s.strip()
    s += "=" * (-len(s) % 4)
    return base64.b64decode(s).decode("utf-8", errors="replace")


def rot_letters(s: str, shift: int = 13) -> str:
    def _rot(c: str) -> str:
        o = ord(c)
        if 65 <= o <= 90:
            return chr((o - 65 + shift) % 26 + 65)
        if 97 <= o <= 122:
            return chr((o - 97 + shift) % 26 + 97)
        return c

    return re.sub(r"[A-Za-z]", lambda m: _rot(m.group(0)), s)


def decode_o(o_value: str, shift: int = 13) -> dict:
    x = b64d(o_value)
    x = b64d(x)
    x = rot_letters(x, shift=shift)
    x = b64d(x)
    return json.loads(x)


def shifter(o_value: str):
    for sh in range(1, 26):
        try:
            data = decode_o(o_value, shift=sh)
            if isinstance(data, dict) and data.get("o"):
                return data
        except Exception:
            pass
    return None


def clean_url(url):
    if "http" not in url:
        url = f"https://{url}"
    parts = urllib.parse.urlsplit(url)
    if not parts.scheme and not parts.netloc:
        return None
    full_url = f"https://{parts.netloc}"
    return full_url


def _search_regex(pattern, string, flags=0, group=None, default=DEFAULT):
    if string is None:
        m = None
    elif isinstance(pattern, (str, re.Pattern)):
        m = re.search(pattern, string, flags)
    else:
        m = None
        for p in pattern:
            m = re.search(p, string, flags)
            if m:
                break
    if m:
        if group is None:
            for g in m.groups():
                if g is not None:
                    return g
            return m.group(0)
        if isinstance(group, (list, tuple)):
            return tuple(m.group(g) for g in group)
        return m.group(group)

    return default


def movie_keys(title: str) -> tuple[str, str]:
    t = (title or "").lower()
    y = YEAR_C.search(t)
    year = y.group(1) if y else ""
    if year:
        t = t.replace(year, " ")
    t = JUNK_C.sub(" ", t)
    t = re.sub(r"[\[\]\(\)\|:]", " ", t)
    t = re.sub(r"[^a-z0-9]+", " ", t)
    t = re.sub(r"\s+", " ", t).strip()

    return t, year


def norm_title(s: str, only_lower=False) -> str:
    if only_lower is True:
        return s.lower()
    s = (s or "").lower()
    s = s.replace("’", "'")
    s = re.sub(r"'s\b", "", s)
    s = re.sub(r"[^a-z0-9]+", " ", s)
    s = re.sub(r"\s+", " ", s).strip()
    return s


def _parse_a_tags(html):
    return re.findall(r"<a[\s\S]+?<\/a>", html)


def token_set_ratio(a: str, b: str) -> int:
    if fuzz:
        return int(fuzz.token_set_ratio(a, b))

    sa = " ".join(sorted(set((a or "").lower().split())))
    sb = " ".join(sorted(set((b or "").lower().split())))
    return int(difflib.SequenceMatcher(None, sa, sb).ratio() * 100)


def _remove_duplicate(data: list, threshold=90):
    kept, keys = [], []
    for d in data:
        k, year = movie_keys(d.get("title", ""))

        def is_dup(prev):
            pk, py = prev
            if year and py and year != py:
                return False
            return token_set_ratio(k, pk) >= threshold

        if any(is_dup(prev) for prev in keys):
            continue

        kept.append(d)
        keys.append((k, year))
    return kept


def _parse_a_tag(s):
    m_href = re.search(r'href\s*=\s*["\']([^"\']+)["\']', s, re.I)
    if not m_href:
        return None, None

    href = unescape(m_href.group(1))

    inner = re.sub(r"<\s*/?\s*a[^>]*>", "", s, flags=re.I)
    inner = re.sub(r"<[^>]+>", " ", inner)
    text = " ".join(inner.split())

    return href, text


def _og_title(html):
    title = _search_regex(r"<title\s*>([^<]+)<", html, default=None)
    if not title:
        title = _search_regex(r'<meta.+title".+content\s*=\s*"([^"]+)"', html)
    title = (title or "").strip()
    return title


def _get_qparam(
    url: str, keys: str | set[str] | list[str] | tuple[str, ...]
) -> str | None:
    if isinstance(keys, str):
        keys = (keys,)

    q = urllib.parse.parse_qs(urllib.parse.urlparse(url).query, keep_blank_values=False)
    for key in keys:
        vals = q.get(key)
        if not vals:
            continue
        v = vals[0]
        if v:
            return urllib.parse.unquote(v)
    return None


def get_tld(url_or_host: str) -> str | None:
    if "://" not in url_or_host:
        url_or_host = "http://" + url_or_host

    host = urllib.parse.urlparse(url_or_host).hostname
    if not host or "." not in host:
        return None

    parts = host.split(".")
    return "." + parts[-1]


def check_both_host_are_same(url: str, new_domain: str) -> str:
    tld1 = get_tld(url)
    tld2 = get_tld(new_domain)
    if tld1 != tld2:
        return True, url.replace(tld1, tld2)
    return False, None


def _og_thumbnail(html):
    thumbnail = _search_regex(r'<meta.+image".+content\s*=\s*"(https?[^"]+)"', html)
    if not thumbnail:
        thumbnail = _search_regex(r'thumbnail[^"]+"\s*:\s*"(http[^"]+)', html)
    return thumbnail


def _parse_m3u8_resolution(stream_inf_line: str):
    m = re.search(r"RESOLUTION=(\d+)x(\d+)", stream_inf_line)
    if not m:
        return None, None
    return int(m.group(1)), int(m.group(2))


def _m3u8_attr(tag_line: str, key: str):
    m = re.search(rf'{re.escape(key)}=("([^"]*)"|[^,]*)', tag_line)
    if not m:
        return None
    val = m.group(2) if m.group(2) is not None else m.group(1)
    return val.strip('"')


def _search_json(pattern, string, default=DEFAULT):
    m = _search_regex(pattern, string, default=default)
    if m:
        return json.loads(m)
    return default


def has_episode_structure(html: str) -> bool:
    return len(re.findall(r"(episode\s*\d+|s\d+\s*e\d+)", html, re.I)) >= 3


def is_series(url: str, html: str) -> bool:
    url_l = (url or "").lower()

    if SERIES_URL_PAT.search(url_l):
        return True
    if re.search(r"full-movie\b|movie\b", url_l, re.I):
        return False

    parts = []
    for pat in (r"<title[^>]*>(.*?)</title>", r"<h1[^>]*>(.*?)</h1>"):
        m = re.search(pat, html, flags=re.I | re.S)
        if m:
            parts.append(m.group(1))

    head = re.sub(r"<[^>]+>", " ", " ".join(parts))
    head = re.sub(r"\s+", " ", head).lower()

    if STRONG_SERIES.search(head):
        return True

    chunk = re.sub(r"<[^>]+>", " ", html[:30000])
    chunk = re.sub(r"\s+", " ", chunk).lower()

    if STRONG_SERIES.search(chunk):
        return True

    if has_episode_structure(chunk):
        return True

    if MOVIE_HINTS.search(head) or MOVIE_HINTS.search(chunk):
        return False

    return False


def url_filename(u: str) -> str | None:
    try:
        path = urllib.parse.urlparse(u).path
        base = os.path.basename(path)
        base = urllib.parse.unquote(base).strip()
        return base or None
    except Exception:
        return None


def determine_filesize(src):
    if not src:
        return {}

    headers = None
    if hasattr(src, "headers"):
        headers = src.headers
    elif isinstance(src, dict):
        headers = src

    if headers:
        cl = headers.get("content-length") or headers.get("Content-Length")
        try:
            n = int(cl)
            if n > 0:
                return {"filesize": n}
        except Exception:
            pass

    text = (
        src.decode("utf-8", "ignore")
        if isinstance(src, bytes)
        else src
        if isinstance(src, str)
        else str(src)
    )

    m = CONTENT_LEN_RE.search(text)
    if m:
        return {"filesize": int(m.group(1))}

    m = BADGE_SIZE_RE.search(text)
    if m:
        num_s, unit = m.groups()
        unit = unit.lower()
        try:
            size = int(float(num_s.replace(",", "")) * _UNITS[unit])
            return {"filesize_approx": size}
        except Exception:
            return {}

    m = SIZE_RE.search(text)
    if m:
        num_s, unit = m.groups()
        try:
            size = int(float(num_s.replace(",", "")) * _UNITS[unit.lower()])
            return {"filesize_approx": size}
        except Exception:
            return {}

    m = INLINE_SIZE_RE.search(text)
    if m:
        num_s, unit = m.groups()
        try:
            size = int(float(num_s.replace(",", "")) * _UNITS[unit.lower()])
            return {"filesize_approx": size}
        except Exception:
            pass

    return {}


def determine_ext(
    url: str | None, other_text: str | None = None, default_ext: str = "mp4"
) -> str:
    def _pick_from(s: str | None) -> str | None:
        if not s:
            return None
        s = str(s).strip()
        s = s.split("?", 1)[0].split("#", 1)[0].rstrip("/")
        if "." not in s:
            return None
        ext = s.rsplit(".", 1)[-1].lower()
        if "google" in url:
            ext = "mkv"
        return ext if ext in KNOWN_EXTENSIONS else None

    return _pick_from(url) or _pick_from(other_text) or default_ext


def bug_reports_message(before=";"):
    msg = (
        f"please report this issue on https://github.com/{REPOSITORY}/issues?q=, "
        f"filling out the appropriate issue template. Confirm you are on the latest version."
    )

    before = (before or "").rstrip()
    if not before or before.endswith((".", "!", "?")):
        msg = msg[:1].upper() + msg[1:]
    return (before + " " if before else "") + msg


def ascii_color(text, code, enable=True):
    return f"\033[{code}m{text}\033[0m" if enable else text


def is_tty(stream):
    try:
        return stream.isatty()
    except Exception:
        return False


class NobiDLError(Exception):
    def __init__(self, msg, *, url=None, expected=False, cause=None, color=True):
        self.url = url
        self.expected = expected
        self.cause = cause
        self.orig_msg = str(msg)
        self.color = color
        super().__init__()
        self.args = (self._msg,)

    @property
    def _msg(self):
        tag = ascii_color("[Error]", "31", self.color and is_tty(sys.stderr))
        where = f": {self.url}" if self.url else ""
        caused = f" (caused by {self.cause!r})" if self.cause else ""
        tail = "" if self.expected else bug_reports_message(";")
        return f"{tag}{where}: {self.orig_msg}{caused}{tail}"

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if name != "args" and hasattr(self, "orig_msg"):
            try:
                self.args = (self._msg,)
            except Exception:
                pass

    def __str__(self):
        return self._msg


def approx_filesize_from_tbr(tbr_kbps, duration):
    if not tbr_kbps or not duration:
        return None
    return int(float(tbr_kbps) * 1000 / 8 * float(duration))


# TODO Remove When needed
# def tbr_from_filesize(filesize_bytes: int, duration_sec: float) -> int | None:
#     if not filesize_bytes or not duration_sec:
#         return {}
#     filesize = (
#         filesize_bytes.get("filesize")
#         or filesize_bytes.get("filesize_approx")
#         or filesize_bytes
#     )
#     if not filesize:
#         return {}
#     tbr = int((filesize * 8) / duration_sec / 1000)
#     return {"tbr": tbr}

# TODO Remove When needed
# def duration_from_anywhere(text: str):
#     if not text:
#         return None

#     m = re.search(r'(?is)"duration"\s*:\s*"(?P<iso>PT[^"]+)"', text)
#     if m:
#         iso = m.group("iso").upper()
#         hh = int(re.search(r"(\d+)H", iso).group(1)) if re.search(r"(\d+)H", iso) else 0
#         mm = int(re.search(r"(\d+)M", iso).group(1)) if re.search(r"(\d+)M", iso) else 0
#         ss = int(re.search(r"(\d+)S", iso).group(1)) if re.search(r"(\d+)S", iso) else 0
#         return hh * 3600 + mm * 60 + ss

#     m = re.search(r"(?<!\d)(\d{1,2}):(\d{2})(?::(\d{2}))?(?!\d)", text)
#     if m:
#         a, b, c = m.group(1), m.group(2), m.group(3)
#         if c is None:
#             return int(a) * 60 + int(b)
#         return int(a) * 3600 + int(b) * 60 + int(c)

#     m = re.search(
#         r"(?is)\b(?:runtime|duration|run\s*time)\b[^0-9]{0,40}((\d{1,2})\s*(?:h|hr|hrs))?[^0-9]{0,10}(\d{1,3})\s*(?:m|min|mins)\b",
#         text,
#     )
#     if m:
#         h = int(m.group(2) or 0)
#         mm = int(m.group(3))
#         return h * 3600 + mm * 60

#     m = re.search(r"(?i)\b(\d{1,4})\s*(?:m|min|mins|minute|minutes)\b", text)
#     if m:
#         return int(m.group(1)) * 60

#     vals = re.findall(r"#EXTINF:([0-9]+(?:\.[0-9]+)?)", text)
#     if vals:
#         return sum(float(x) for x in vals)
#     if "#EXT-X-TARGETDURATION" in text:
#         td = re.search(r"#EXT-X-TARGETDURATION:(\d+)", text)
#         segs = sum(
#             1 for ln in text.splitlines() if ln.strip() and not ln.startswith("#")
#         )
#         if td and segs:
#             return int(td.group(1)) * segs

#     return None


def fix_entries(self, entries: list) -> list:
    fixed = {}

    for entry in entries:
        season = entry.get("season")
        episode = entry.get("episode")

        if not all((season, episode)):
            self.show_warning("Missing season/episode, skipping")
            continue

        key = (season, episode)

        if key not in fixed:
            fixed[key] = entry
        else:
            fixed[key]["formats"].extend(entry.get("formats", []))

    return list(fixed.values())


def random_id() -> str:
    return str(random.randint(1_000_000, 9_999_999))[:4]


def filename_from_title(title: str, url, ext=None) -> str:
    ext = determine_ext(url, title) if ext is None else ext
    norm_title, _ = movie_keys(title)
    title_ext = determine_ext(norm_title, norm_title)
    if title_ext:
        norm_title = norm_title.replace(title_ext, "")
    filename = f"{norm_title}.{ext}" if "." not in ext else norm_title + ext
    return filename



from .common import BestFMT
from .fmt_resolver import Resolve_FMTS
from .hls import HlsDL
from .http import HttpDownloader


class Downloader:
    def __init__(self, md, info_dict, options, logger):
        self.info_dict = info_dict
        self.logger = logger
        self.options = options
        self.md = md

    def __call__(self):
        return self.downloader()

    def downloader(self):
        formats = self.info_dict.get("formats") or []
        bfmt = BestFMT(formats)()
        bfmt = Resolve_FMTS(self.md, self.info_dict, bfmt, None)()
        url = bfmt.get("url")
        if "m3u8" in url:
            return HlsDL(bfmt, self.info_dict, self.options, self.logger)()
        else:
            return HttpDownloader(bfmt, self.info_dict, self.options, self.logger)()



import itertools
import re

from ..nobi_dl import ExtractorBase
from ..utils import (
    _og_thumbnail,
    _og_title,
    _parse_a_tag,
    _parse_a_tags,
    _parse_resolution,
    _search_regex,
    determine_ext,
    determine_filesize,
    fix_entries,
    is_series,
    random_id,
)


class MoviesModME(ExtractorBase):
    _IE_NAME = "Moviesmod"
    _VALID_URL = r"https://?moviesmod\.[^/]+/[^?&#]+"
    _SEARCH = True

    @property
    def host(self):
        return self.host_finder.moviemod()

    def parse_content(self, html):
        results = []
        for article in re.findall(r"<article[\s\S]+?<\/article>", html):
            href = _search_regex(r'href="([^"]+)"', article)
            href = f"{self.host}{href}" if "https" not in href else href
            thumbnail = _search_regex(r'src="([^"]+)"', article)
            thumbnail = (
                f"{self.host}{thumbnail}" if "https" not in thumbnail else thumbnail
            )
            title = _search_regex(r'title="([^"]+)"', article)
            results.append(
                {
                    "title": title,
                    "thumbnail": thumbnail,
                    "url": href,
                    "quality": _parse_resolution(title),
                    "source": "hdhub",
                }
            )
        return results

    def _real_search(self, query, limit=50):
        search_result = []
        for pagenum in itertools.count(1):
            page = self._download_webpage(
                f"{self.host}/search/{query}/page/{pagenum}",
                headers={"referer": f"{self.host}/"},
            )
            search_result.extend(self.parse_content(page))
            is_next = _search_regex(
                r'(?is)<a\s*class\s*=\s*"next[^>]+>\s*next\s*?(?:[^<]+)?<',
                page,
                default=False,
            )
            if is_next is False or search_result >= limit:
                break

        return search_result

    def _get_movie(self, url_or_webpage):
        if url_or_webpage.startswith("http") or "<html" not in url_or_webpage:
            webpage = self._download_webpage(url_or_webpage)
        else:
            webpage = url_or_webpage

        formats = []
        for h4 in re.findall(r"<h4[\s\S]+?<\/h4>[\s\S]+?<a[\s\S]+?<\/a>", webpage):
            h4_texts = _search_regex(r"(<h4[\s\S]+?<\/h4>)", h4)
            for atag in _parse_a_tags(h4):
                href, name = _parse_a_tag(atag)
                label = (name or "").lower()
                if not any(
                    k in label for k in ("download links", "download link", "d links")
                ):
                    continue
                formats.append(
                    {
                        "format_id": random_id(),
                        "url": href,
                        "acodec": True,
                        "vcodec": True,
                        "ext": determine_ext(href, href),
                        **determine_filesize(h4_texts),
                        **_parse_resolution(h4_texts),
                    }
                )

        return {
            "title": _og_title(webpage),
            "thumbnail": _og_thumbnail(webpage),
            "formats": formats,
        }

    def extract_h3_text(self, block):
        h3_text = _search_regex(r"<h3[\s\S]+?<\/h3>", block)
        h3_text = re.sub(r"<[^>]+", " ", h3_text)
        return h3_text

    def extract_episodes(self, url, h3_text, op_webpage):
        webpage = self._download_webpage(url)
        episodes = []
        op_title = _og_title(op_webpage)
        season = _search_regex(r"(?is)(?:season|s0|s)\s*(\d+)", h3_text)
        for atag in _parse_a_tags(webpage):
            href, name = _parse_a_tag(atag)
            label = (name or "").lower()
            if not any(k in label for k in ("episode", "episo", "sode")):
                continue
            if not any(k in href for k in ("modpro.blog", "unblockedgames.world")):
                continue
            if any(k in label for k in ("all episodes", "episodes batch", "batch")):
                continue
            if "episodes" in href:
                continue
            ep_code = _search_regex(r"(?is)(?:episode|ep|ep0)\s*(\d+)", name)
            episodes.append(
                {
                    "title": f"{op_title} - season {season} - ep{ep_code}",
                    "season": season,
                    "episode": ep_code,
                    "formats": [
                        {
                            "url": href,
                            "acodec": True,
                            "vcodec": True,
                            "ext": "mkv",
                            **determine_filesize(h3_text),
                            **_parse_resolution(h3_text),
                        }
                    ],
                }
            )

        return episodes

    def entries(self, block, h3tag, op_webpage):
        h3tag = h3tag if h3tag else self.extract_h3_text(block)
        entires = []
        for atag in _parse_a_tags(block):
            href, name = _parse_a_tag(atag)
            label = (name or "").lower()
            if not any(k in label for k in ("episode", "episode links")):
                continue
            entires.extend(self.extract_episodes(href, h3tag, op_webpage))

        return entires

    def _get_series(self, url_or_webpage, url):
        if url_or_webpage.startswith("http") or "<html" not in url_or_webpage:
            webpage = self._download_webpage(url_or_webpage)
        else:
            webpage = url_or_webpage
        entries = []
        blocks = re.findall(r"<h3[\s\S]+?<\/h3>[\s\S]+?<\/p>", webpage)
        for block in blocks:
            block = re.sub(r"<h3[\s\S]+?<\s*hr\s*>", "", block)
            if "<a" not in block:
                continue
            h3_text = self.extract_h3_text(block)
            if not any(k in h3_text.lower() for k in ("season", "s01", "s1")):
                continue
            entries.extend(self.entries(block, h3_text, webpage))

        entries = fix_entries(self, entries)
        return self.playlist_result(entries, url)

    def _real_extract(self, url):
        webpage = self._download_webpage(url)
        if is_series(url, webpage):
            return self._get_series(webpage, url)
        else:
            return self._get_movie(webpage)



from typing import get_args

import curl_cffi
from curl_cffi import const, requests
from curl_cffi import exceptions as req_exc
from curl_cffi.requests.impersonate import DEFAULT_CHROME, BrowserTypeLiteral

SESSION = requests.Session()
DEFAULT_TIMEOUT = 20


class ImpersonateTargets:
    def __init__(self, logger):
        self.to_screen = logger.to_screen

    @staticmethod
    def _label(name: str, ver: str, only_client=False) -> str:
        if only_client is True:
            return name
        ver = (ver or "").strip().lstrip("_").replace("_", ".")
        return f"{name}-{ver}" if ver else name

    def _de_label(self, target):
        pass

    def _de_label(self, target: str | None) -> str | None:
        if not target:
            return None

        t = target.strip().lower()

        if ":" in t:
            t = t.split(":", 1)[0]

        t = t.replace(" ", "")
        t = t.replace("-ios", "_ios").replace("_ios", "_ios")
        t = t.replace("-android", "_android")
        t = t.replace("-beta", "_beta")

        if "-" in t:
            name, ver = t.split("-", 1)
        else:
            name, ver = t, ""

        if name == "safari":
            ver = ver.replace(".", "_")

        if name == "tor":
            ver = ver.replace(".", "")

        if name in {"chrome", "firefox", "edge"}:
            ver = ver.replace(".", "").replace("_", "")

        return f"{name}{ver}" if ver else name

    @property
    def all_targets(self):
        return list(get_args(BrowserTypeLiteral))

    def valid_targets(self):
        all_targets = self.all_targets
        filtered_targets = []
        for target in all_targets:
            client, _ = self._parse_target(target)
            if not client:
                continue
            filtered_targets.append(client)
        return filtered_targets

    def _parse_target(self, t: str, only_client=False) -> tuple[str, str]:
        if t.startswith("chrome"):
            ver = t.removeprefix("chrome")
            if t.endswith("_android"):
                return None, None
            return (self._label("Chrome", ver, only_client), "Macos-14")

        if t.startswith("firefox"):
            ver = t.removeprefix("firefox")
            return (self._label("Firefox", ver, only_client), "Macos-14")

        if t.startswith("safari"):
            ver = t.removeprefix("safari")
            if "ios" in t:
                ver = ver.replace("_ios", "")
                return (self._label("Safari", ver), "Ios-18.4")
            return (self._label("Safari", ver, only_client), "Macos-15")

        if t.startswith("edge"):
            ver = t.removeprefix("edge")
            return (self._label("Edge", ver, only_client), "Windows-10")

        if t.startswith("tor"):
            ver = t.removeprefix("tor")
            v = f"{ver[:2]}.{ver[2:]}" if ver.isdigit() and len(ver) == 3 else ver
            return (self._label("Tor", v, only_client), "Macos-14")

    def show_impersonate_table(self) -> int:
        rows = []
        for t in self.all_targets:
            client, os = self._parse_target(t)
            if not client:
                continue
            rows.append((client, os, "curl_cffi"))

        rows.sort(key=lambda r: (r[0].split("-")[0], r[0]))

        self.to_screen("[info] Available impersonate targets")
        self.to_screen(f"{'Client':<12}  {'OS':<12}  {'Source'}")
        self.to_screen("-" * 36)
        for client, os, src in rows:
            self.to_screen(f"{client:<12}  {os:<12}  {src}")
        return 0


class Curl_cfiiRH:
    def __init__(
        self,
        url,
        method="GET",
        impersonate=None,
        headers=None,
        cookies=None,
        query=None,
        data=None,
        timeout=DEFAULT_TIMEOUT,
        retries=2,
        verbose=False,
        logger=None,
        stream=False,
    ):
        self.logger = logger
        self.verbose = verbose
        self.url = url
        self.method = method
        self.impersonate = impersonate
        self.headers = headers or {}
        self.cookies = cookies or {}
        self.query = query or {}
        self.data = data or {}
        self.timeout = timeout
        self.retries = retries
        self.stream = stream

    def __call__(self):
        return self._request()

    def _print_verbose(self, msg):
        if self.verbose is True:
            return self.logger.to_stdout(msg)

    def print_traffic(self) -> dict:
        if not self.verbose:
            return {}
        self.logger.to_stdout(f"[info] Sending Request via {curl_cffi.__name__}")

        return {const.CurlOpt.VERBOSE: 1}

    def _request(self):
        from ._request_handler import std_headers

        default_headers = dict(std_headers)
        default_headers.update(self.headers)
        curl_opts = self.print_traffic()
        if curl_opts:
            SESSION.curl_options = curl_opts

        if self.impersonate is None:
            self.impersonate = DEFAULT_CHROME

        last_err = None
        for _ in range(self.retries + 1):
            try:
                resp = SESSION.request(
                    self.method,
                    self.url,
                    cookies=self.cookies,
                    params=self.query,
                    data=self.data,
                    timeout=self.timeout,
                    headers=default_headers,
                    impersonate=self.impersonate,
                    stream=self.stream,
                )

                SESSION.cookies.update(resp.cookies)
                if resp.status_code in (429, 503, 502, 504):
                    last_err = RuntimeError(f"HTTP {resp.status_code}")
                    continue

                return resp

            except req_exc.Timeout as e:
                last_err = e
                continue

        raise last_err



import http.client

import requests
from requests import exceptions as req_exc

from ._curl_cffi import Curl_cfiiRH

GLOBAL_SESSION = requests.Session()


class RequestsRH:
    def __init__(
        self,
        url,
        headers=None,
        cookies=None,
        query=None,
        data=None,
        timeout=20,
        method=None,
        retries=2,
        verbose=False,
        logger=None,
        stream=False,
    ):
        self.logger = logger
        self.verbose = verbose
        self.url = url
        self.headers = headers or {}
        self.cookies = cookies or {}
        self.query = query or {}
        self.data = data or {}
        self.timeout = timeout or 20
        self.method = method or "GET"
        self.retries = retries
        self.stream = stream

    def __call__(self):
        return self._request()

    def _print_verbose(self, msg):
        if self.verbose is True:
            return self.logger.to_stdout(msg)

    def print_traffic(self):
        if self.verbose is False:
            return
        self._print_verbose(f"Sending Request via {requests.__name__}")
        http.client.HTTPSConnection.debuglevel = 1
        http.client.HTTPConnection.debuglevel = 1

    def _request(self):
        from ._request_handler import std_headers

        self.print_traffic()
        default_headers = dict(std_headers)
        if self.headers:
            default_headers.update(self.headers)

        last_err = None
        for _ in range(self.retries + 1):
            try:
                resp = GLOBAL_SESSION.request(
                    self.method,
                    self.url,
                    cookies=self.cookies,
                    params=self.query,
                    data=self.data,
                    timeout=self.timeout,
                    headers=default_headers,
                    stream=self.stream,
                )
                GLOBAL_SESSION.cookies.update(resp.cookies)
                if resp.status_code in (429, 503, 502, 504):
                    last_err = RuntimeError(f"HTTP {resp.status_code}")
                    continue

                return resp

            except req_exc.Timeout as e:
                last_err = e
                continue

            except req_exc.ConnectionError:
                self.logger.to_stdout("[network] Connection Error Forcing to Curl cffi")
                return Curl_cfiiRH(
                    self.url,
                    self.method,
                    None,
                    self.headers,
                    self.cookies,
                    self.query,
                    self.data,
                    self.timeout,
                    self.retries,
                    self.verbose,
                    self.logger,
                    self.stream,
                )()

        raise last_err



from .common import ExtractorBase


class Searcher(ExtractorBase):
    def __init__(self, md=None, force_extractor=None):
        super().__init__(md)
        self.md = md
        self.force_extractor = force_extractor

    def get_all_searcher(self) -> list[dict]:
        searchers = []

        for extractor in self.gen_extractors(self.md):
            if not hasattr(extractor, "_real_search"):
                continue

            ie_name = getattr(extractor, "_IE_NAME", None)
            if not ie_name:
                continue

            searchers.append(
                {
                    "ie": extractor,
                    "name": ie_name,
                }
            )

        return searchers

    def search(self, query: str) -> dict:
        raw_result = self._raw_search(query)

        if not raw_result:
            return {
                "query": query,
                "total": 0,
                "best": None,
                "results": [],
            }

        ranked = self._rank_results(raw_result, query)

        return {
            "query": query,
            "total": len(ranked),
            "best": ranked[0] if ranked else None,
            "results": ranked,
        }

    def _rank_results(self, results, query):
        q = query.lower()

        def score(r):
            s = 0
            title = r.get("title", "").lower()

            if q in title:
                s += 50

            quality = r.get("quality") or []
            if "1080p" in quality:
                s += 20
            if "2160p" in quality:
                s += 30

            return s

        return sorted(results, key=score, reverse=True)

    def _raw_search(self, query: str):
        searchers = self.get_all_searcher()

        if self.force_extractor:
            searchers = [
                s
                for s in searchers
                if self.force_extractor.lower() in s["ie"].__class__.__name__.lower()
                or self.force_extractor.lower() in s["name"].lower()
            ]

            if not searchers:
                self.show_warning(
                    "The extractor doesn't support search or is not implemented yet. "
                    "Try --list-extractors"
                )
                return []

        results = []

        for s in searchers:
            ie = s["ie"]

            if getattr(ie, "_SEARCH", True) is False:
                self.show_warning(f"skipping {s['name']} (search disabled)")
                continue
            self.logger.set_ie(ie, "search")
            try:
                res = ie._real_search(query)
                if res:
                    results.extend(res)
            except Exception as e:
                self.write_debug(f"Search failed for {s['name']}: {e}")

        return results



#!/usr/bin/env python3

import sys

if __package__ is None and not getattr(sys, "frozen", False):
    import os.path

    path = os.path.realpath(os.path.abspath(__file__))
    sys.path.insert(0, os.path.dirname(os.path.dirname(path)))

from nobi_dl import main

if __name__ == "__main__":
    main()



import re
import types

ANSI_RE = re.compile(r"\x1b\[[0-9;]*m")


def remove_terminal_sequences(s: str) -> str:
    return ANSI_RE.sub("", s or "")


class Namespace(types.SimpleNamespace):
    def __iter__(self):
        return iter(self.__dict__.values())

    @property
    def items_(self):
        return self.__dict__.items()


Styles = Namespace(
    HEADERS="yellow",
    EMPHASIS="light blue",
    FILENAME="green",
    ID="green",
    DELIM="blue",
    ERROR="red",
    BAD_FORMAT="light red",
    WARNING="yellow",
    SUPPRESS="light black",
)
ALLOW_COLORS = True
OUT_HANDLE = types.SimpleNamespace(encoding="utf-8")

CONTROL_SEQUENCES = {
    "DOWN": "\n",
    "UP": "\033[A",
    "ERASE_LINE": "\033[K",
    "RESET": "\033[0m",
}

_COLORS = {
    "BLACK": "0",
    "RED": "1",
    "GREEN": "2",
    "YELLOW": "3",
    "BLUE": "4",
    "PURPLE": "5",
    "CYAN": "6",
    "WHITE": "7",
}

_TEXT_STYLES = {
    "NORMAL": "0",
    "BOLD": "1",
    "UNDERLINED": "4",
}


def format_text(text, f):
    f = f.upper()
    tokens = f.strip().split()

    bg_color = ""
    if "ON" in tokens:
        if tokens[-1] == "ON":
            raise SyntaxError(f"Empty background format specified in {f!r}")
        if tokens[-1] not in _COLORS:
            raise SyntaxError(f"{tokens[-1]} in {f!r} must be a color")
        bg_color = f"4{_COLORS[tokens.pop()]}"
        if tokens[-1] == "LIGHT":
            bg_color = f"0;10{bg_color[1:]}"
            tokens.pop()
        if tokens[-1] != "ON":
            raise SyntaxError(f"Invalid format {f.split(' ON ', 1)[1]!r} in {f!r}")
        bg_color = f"\033[{bg_color}m"
        tokens.pop()

    if not tokens:
        fg_color = ""
    elif tokens[-1] not in _COLORS:
        raise SyntaxError(f"{tokens[-1]} in {f!r} must be a color")
    else:
        fg_color = f"3{_COLORS[tokens.pop()]}"
        if tokens and tokens[-1] == "LIGHT":
            fg_color = f"9{fg_color[1:]}"
            tokens.pop()
        fg_style = tokens.pop() if tokens and tokens[-1] in _TEXT_STYLES else "NORMAL"
        fg_color = f"\033[{_TEXT_STYLES[fg_style]};{fg_color}m"
        if tokens:
            raise SyntaxError(f"Invalid format {' '.join(tokens)!r} in {f!r}")

    if fg_color or bg_color:
        text = text.replace(CONTROL_SEQUENCES["RESET"], f"{fg_color}{bg_color}")
        return f"{fg_color}{bg_color}{text}{CONTROL_SEQUENCES['RESET']}"
    else:
        return text


def _format_text(handle, allow_colors, text, f, fallback=None, *, test_encoding=False):
    text = str(text)
    if test_encoding:
        original_text = text
        encoding = getattr(handle, "encoding", None) or "ascii"
        text = text.encode(encoding, "ignore").decode(encoding)
        if fallback is not None and text != original_text:
            text = fallback
    return (
        format_text(text, f)
        if allow_colors is True
        else text
        if fallback is None
        else fallback
    )


def _format_out(text, f, fallback=None, *, test_encoding=False):
    return _format_text(
        OUT_HANDLE, ALLOW_COLORS, text, f, fallback, test_encoding=test_encoding
    )


def join_nonempty(*items, delim=", ", inner_delim=" "):
    if len(items) == 0:
        return ""
    out = []
    for it in items:
        if not it:
            continue
        if isinstance(it, (list, tuple)):
            out.extend([x for x in it if x])
        else:
            out.append(it)
    return (delim if delim is not None else inner_delim).join(out) if out else ""


def _list_format_headers(*headers):
    if headers:
        return [_format_out(header, Styles.HEADERS) for header in headers]
    return headers


def shorten_protocol_name(fmt: dict) -> str:
    url = fmt.get("url")
    p = "https" if "m3u8" not in url else "m3u8"
    return {
        "https": "http",
        "m3u8": "m3u8",
    }.get(p, p)


def format_decimal_suffix(num):
    if num is None:
        return None
    try:
        n = float(num)
    except Exception:
        return str(num)
    for unit, div in (("G", 1e9), ("M", 1e6), ("k", 1e3)):
        if abs(n) >= div:
            v = n / div
            return f"{v:.0f}{unit}" if v >= 10 else f"{v:.1f}{unit}"
    return f"{n:.0f}"


def format_bytes(num):
    if num is None:
        return None
    try:
        n = float(num)
    except Exception:
        return str(num)
    units = ["B", "KiB", "MiB", "GiB", "TiB"]
    i = 0
    while n >= 1024 and i < len(units) - 1:
        n /= 1024
        i += 1
    return f"{n:.1f}{units[i]}" if i else f"{int(n)}{units[i]}"


def filesize_from_tbr(tbr_kbps, duration):
    if not tbr_kbps or not duration:
        return None
    try:
        return int(float(tbr_kbps) * 1000 / 8 * float(duration))
    except Exception:
        return None


def format_resolution(f):
    if f.get("resolution"):
        return f["resolution"]
    w, h = f.get("width"), f.get("height")
    if w and h:
        return f"{w}x{h}"
    return ""


def format_field(f: dict, key=None, fmt="%s", default="", func=None, ignore=()):
    if callable(key) and func is None:
        func = key
        key = None

    val = None
    if key is None:
        val = f
    elif isinstance(key, str):
        val = f.get(key)
    else:
        val = key(f)

    if val in ignore:
        return ""
    if val is None:
        return default if default is not None else ""

    if func:
        try:
            val = func(val)
        except Exception:
            pass

    try:
        return fmt % val if fmt else str(val)
    except Exception:
        return str(val)


def simply_type(f: dict) -> str:
    ext = (f.get("ext") or "").lower()
    if ext == "zip":
        return "zipped"

    if ext in ("mp4", "mkv", "webm", "avi", "mov"):
        return "direct"

    acodec = f.get("acodec")
    vcodec = f.get("vcodec")

    if acodec is None and vcodec is None:
        return "unknown"

    if acodec and vcodec:
        return "premerged"
    if acodec and not vcodec:
        return "audio only"
    if vcodec and not acodec:
        return "video only"
    return "unknown"


def render_table(header_row, data, delim=False, extra_gap=0, hide_empty=False):
    vbar = _format_out("│", Styles.DELIM, "|", test_encoding=True)

    def width(s):
        return len(remove_terminal_sequences(str(s)).replace("\t", ""))

    def normalize_row(row, n):
        row = list(row)
        if len(row) < n:
            row += [""] * (n - len(row))
        return row[:n]

    if not header_row:
        return ""

    ncols = len(header_row)
    data = [normalize_row(r, ncols) for r in (data or [])]
    header_row = normalize_row(header_row, ncols)

    def get_max_lens(rows):
        cols = list(zip(*rows)) if rows else [[] for _ in range(ncols)]
        return [max((width(v) for v in col), default=0) for col in cols]

    if hide_empty:
        max_lens = get_max_lens(data)
        keep = [ml > 0 for ml in max_lens]
        header_row = [c for k, c in zip(keep, header_row) if k]
        data = [[c for k, c in zip(keep, r) if k] for r in data]
        ncols = len(header_row)
        data = [normalize_row(r, ncols) for r in data]

    rows = [header_row, *data]
    max_lens = get_max_lens(rows)
    extra_gap = max(1, extra_gap)

    def pad_row(row):
        out = []
        for i, text in enumerate(map(str, row)):
            if "\t" in text:
                out.append(text.replace("\t", " " * (max_lens[i] - width(text))))
            else:
                out.append(text + " " * (max_lens[i] - width(text)))
        return out

    header = pad_row(header_row)
    body = [pad_row(r) for r in data]

    lines = [
        (" " * extra_gap).join(header[:1] + sum(([vbar, c] for c in header[1:]), []))
    ]

    if delim:
        hline = _format_out("─", Styles.DELIM, "-", test_encoding=True)
        sep = hline * (len(remove_terminal_sequences(lines[0])))
        lines.append(sep)

    for r in body:
        lines.append(
            (" " * extra_gap).join(r[:1] + sum(([vbar, c] for c in r[1:]), []))
        )

    return "\n".join(s.rstrip() for s in lines)


def render_formats_table(formats, info_dict=None):
    if not formats:
        return ""

    info_dict = info_dict or {}

    table = [
        [
            _format_out(format_field(f, "format_id"), Styles.ID),
            format_field(f, "ext"),
            format_resolution(f),
            format_field(f, "fps", "\t%d", func=round),
            (
                format_field(f, "filesize", " \t%s", func=format_bytes)
                or format_field(f, "filesize_approx", "≈\t%s", func=format_bytes)
                or format_field(
                    filesize_from_tbr(f.get("tbr"), info_dict.get("duration")),
                    None,
                    _format_out("~\t%s", Styles.SUPPRESS),
                    func=format_bytes,
                )
            ),
            format_field(f, "tbr", "\t%dk", func=round),
            shorten_protocol_name(f),
            simply_type(f),
            join_nonempty(
                format_field(f, "language", "[%s]"),
                join_nonempty(
                    format_field(f, "format_note"),
                    format_field(f, "container", ignore=(None, f.get("ext"))),
                    delim=", ",
                ),
                delim=" ",
            ),
        ]
        for f in formats
        if f.get("ext") and f.get("url")
    ]

    header_line = _list_format_headers(
        "ID",
        "EXT",
        "RESOLUTION",
        "\tFPS",
        "\tFILESIZE",
        "\tTBR",
        "PROTO",
        "TYPE",
        "MORE INFO",
    )

    return render_table(
        header_line,
        table,
        delim=_format_out("─", Styles.DELIM, "-", test_encoding=True),
        hide_empty=True,
    )



import random

from ..utils import NobiDLError
from ._urllib import UrllibRH


def random_user_agent():
    USER_AGENT_TMPL = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/{} Safari/537.36"
    CHROME_MAJOR_VERSION_RANGE = (137, 143)
    return USER_AGENT_TMPL.format(
        f"{random.randint(*CHROME_MAJOR_VERSION_RANGE)}.0.0.0"
    )


std_headers = {
    "User-Agent": random_user_agent(),
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
    "Accept-Language": "en-us,en;q=0.5",
    "Sec-Fetch-Mode": "navigate",
}


class RequestPrefrence:
    def __call__(self):
        return self._handler_prefrenece()

    def _handler_prefrenece(self):
        try:
            from ._request import RequestsRH

            return RequestsRH
        except ImportError:
            pass

        try:
            from ._curl_cffi import Curl_cfiiRH

            return Curl_cfiiRH
        except ImportError:
            pass

        return UrllibRH


class Request:
    def __init__(
        self,
        url,
        headers=None,
        cookies=None,
        query=None,
        data=None,
        timeout=20,
        method="GET",
        retries=3,
        verbose=False,
        logger=None,
        impersonate=None,
        stream=False,
    ):
        self.logger = logger
        self.verbose = verbose
        self.url = url
        self.cookies = cookies or {}
        self.headers = headers or {}
        self.query = query or {}
        self.data = data or {}
        self.timeout = timeout or 20
        self.method = method or "GET"
        self.retries = retries
        self.impersonate = impersonate
        self.stream = stream

    def __call__(self):
        return self._request()

    def _client_cls(self):
        return RequestPrefrence()()

    def _request(self):
        Handler = self._client_cls()

        kw = dict(
            url=self.url,
            verbose=self.verbose,
            logger=self.logger,
            headers=self.headers,
            cookies=self.cookies,
            query=self.query,
            data=self.data,
            timeout=self.timeout,
            method=self.method,
            retries=self.retries,
            stream=self.stream,
        )

        if Handler == "Curl_cfiiRH":
            from ._curl_cffi import ImpersonateTargets

            it = ImpersonateTargets(self.logger)
            if self.impersonate:
                imp = it._de_label(self.impersonate)
                if imp not in it.all_targets:
                    raise NobiDLError(
                        f'Impersonate target "{self.impersonate}" is not available.'
                        f" See --list-impersonate-targets for available targets."
                        f" This request requires browser impersonation, however you may be missing dependencies"
                        f" required to support this target."
                    )
                kw["impersonate"] = imp

        return Handler(**kw)()



import re

from nobi_dl import ExtractorBase
from nobi_dl.utils import (
    _og_thumbnail,
    _og_title,
    _parse_a_tag,
    _parse_a_tags,
    _parse_resolution,
    _search_regex,
    determine_ext,
    determine_filesize,
    fix_entries,
    is_series,
    random_id,
)

from .hdhub import Hdhub4uME


class FourkhdhubME(ExtractorBase):
    _IE_NAME = r"4khdhub"
    _VALID_URL = r"https://?4khdhub\.[^/]+/[^?&#]+"
    _SEARCH = True

    @property
    def HOST(self):
        return self.host_finder._4k_host()

    @property
    def hdhubme(self):
        return Hdhub4uME(self.md)

    def _clean_meta(self, meta: str) -> str:
        meta = (meta or "").strip()
        meta = re.compile(r"\s+").sub(" ", meta)
        meta = re.compile(r"\s*•\s*").sub(" • ", meta)
        return meta

    def _parse_search_4k_atags(self, atag: str, host: str):
        title = _search_regex(
            r'class="[^"]*movie[^"]*title[^"]*"\s*>\s*([^<]+?)\s*<', atag
        ).strip()
        meta = _search_regex(
            r'class="[^"]*movie[^"]*meta[^"]*"\s*>\s*([^<]+?)\s*<', atag
        ).strip()
        if meta:
            meta = self._clean_meta(meta)
        thumbnail = _search_regex(
            r'\bsrc="([^"]+)"[^>]*\bclass="[^"]*cover[^"]*"', atag
        ).strip()
        link = _search_regex(r'<a[^>]+href="([^"]+)"', atag).strip()
        year = _search_regex(r"\b(19\d{2}|20\d{2})\b", meta, default="").strip()
        if not link.startswith("http"):
            link = f"{host}{link}"

        return {
            "title": f"{title} - {meta}",
            "url": link,
            "thumbnail": thumbnail,
            "year": year,
            "quality": _parse_resolution(title),
            "source": "4khub",
        }

    def _real_search(self, query):
        webpage = self._download_webpage(f"{self.HOST}/?s={query}")
        results = []
        for atag in re.findall(
            r'(<a\s*href\s*=[^>]+class\s*="movie[^"]+"\s*>[\s\S]+?<\/a>)', webpage
        ):
            results.append(self._parse_search_4k_atags(atag, self.HOST))
        return results

    def _extract_formats(self, webpage):
        formats = []
        for filetag in re.findall(
            r'<div\s*class\s*=s*"file-title[^>]+>[\s\S]+?<\/div>\s*?<\/div>', webpage
        ):
            filelabel = _search_regex(r"\bfile-title[^>]+>([^<]+)<", filetag)
            filelabel_div = _search_regex(
                r"<div[^>]+\bfile-title[^>]+>[\s\S]+?<\/div>\s*[\s\S]+?<\/div>", filetag
            )
            for atag in _parse_a_tags(filetag):
                href, name = _parse_a_tag(atag)
                label = (name or "").lower()
                if not any(k in label for k in ("hubcloud", "download hubcloud")):
                    continue
                formats.append(
                    {
                        "format_id": random_id(),
                        "url": href,
                        "ext": determine_ext(filelabel, filelabel),
                        "acodec": True,
                        "vcodec": True,
                        **determine_filesize(filelabel_div),
                        **_parse_resolution(filelabel),
                    }
                )
        return formats

    def _get_movie(self, url_or_webpage):
        if "<html" in url_or_webpage:
            webpage = url_or_webpage
        else:
            webpage = self._download_webpage(url_or_webpage)

        w1, w2 = re.search(
            r"<span.*(?:Audios|audios)(?:[^\w]+(?P<w1>\w+\s*[^<]+))?(?:[^>]+>(?P<w2>[^<]+))",
            webpage,
        ).groups()
        languages = w2 if "span" in w1 else w1
        return {
            "title": _og_title(webpage),
            "languages": languages,
            "thumbnail": _og_thumbnail(webpage),
            "formats": self._extract_formats(webpage),
        }

    def entries(self, webpage):
        entries = []

        for season in re.findall(
            r'<div\s*class\s*=\s*"episode-download-item"?[^>]+?>[\s\S]+?<\/div>\s*<\/div>',
            webpage,
        ):
            format_id = _search_regex(r"\bfile-title[^>]+>([^<]+)<", season).strip()

            ep = _search_regex(r"(?is)(?:s\d+\s*)?(?:episode|ep|e)\s*(\d+)", format_id)
            season_num = _search_regex(r"(?is)(?:season|s0|s)\s*(\d+)", season) or "1"

            size = determine_filesize(season)
            ext = determine_ext(format_id)

            for atag in _parse_a_tags(season):
                href, name = _parse_a_tag(atag)
                label = (name or "").lower()

                if not any(k in label for k in ("hubcloud", "download hubcloud")):
                    continue

                entries.append(
                    {
                        "title": f"{format_id} - season {season_num} - ep {ep}",
                        "season": season_num,
                        "episode": ep,
                        "formats": [
                            {
                                "format_id": random_id(),
                                "url": href,
                                "acodec": True,
                                "vcodec": True,
                                "ext": ext,
                                **size,
                                **_parse_resolution(season),
                            }
                        ],
                    }
                )

        return entries

    def _get_series(self, url_or_webpage, url):
        if "<html" in url_or_webpage:
            webpage = url_or_webpage
        else:
            webpage = self._download_webpage(url_or_webpage)

        entires = fix_entries(self, self.entries(webpage))
        return self.playlist_result(entires, url, _og_title(webpage))

    def _real_extract(self, url):
        webpage = self._download_webpage(url)
        if is_series(url, webpage):
            return self._get_series(webpage, url)
        else:
            return self._get_movie(webpage)



from .bollyflix import BollyFlixME
from .fourkhdhub import FourkhdhubME
from .hdhub import Hdhub4uME
from .moviesmod import MoviesModME
from .vegamovies import VegaMoviesME

ALL = [BollyFlixME, Hdhub4uME, VegaMoviesME, FourkhdhubME, MoviesModME]



import json
import platform
import ssl
import struct
import sys
import time

from nobi_dl.format_render import render_formats_table
from nobi_dl.options import options

from .common import ExtractorBase
from .downloader import Downloader, Resolve_FMTS
from .logger import Log
from .search import Searcher
from .utils import (
    NobiDLError,
    ascii_color,
)
from .version import __version__


class NobiDL:
    def __init__(self, opts, args, opt_error):
        self.args = args
        self.opts = opts
        self.impersonate = getattr(opts, "impersonate", None)
        self.logger = Log(opts)
        self.write_debug = self.logger.write_debug
        self.to_screen = self.logger.to_screen
        self.raise_error = self.logger.error
        self.show_warning = self.logger.warning
        self.write = self.logger.to_stdout
        self.list_impersonate = self.opts.list_impersonate
        self.dump_json = self.opts.dump_json
        self.downloader = Downloader
        self.verbose = self.opts.verbose
        self.extractor_base = ExtractorBase(self)
        self.opt_error = opt_error
        self.list_extractors = self.opts.list_extractors
        self.force_extractor = self.opts.force_extractor
        self.searcher = Searcher(self, self.force_extractor)
        self._ie = None

    def _real_initialize(self):
        if self.list_impersonate is True:
            return 0
        argv = sys.argv[1:]
        pyver = sys.version.split()[0]
        impl = platform.python_implementation()
        arch = "x86_64" if struct.calcsize("P") * 8 == 64 else "x86"
        plat = platform.platform()
        openssl = getattr(ssl, "OPENSSL_VERSION", "unknown")
        ies = ExtractorBase.gen_extractors(self)
        request_handlers = ", ".join(self.request_handlers())
        self.write_debug(f"Command-line config: {argv}")
        self.write_debug(f"nobi-dl version: {__version__}")
        self.write_debug(f"Python {pyver} ({impl} {arch}) - {plat} ({openssl})")
        self.write_debug(f"exe: {sys.executable}")
        self.write_debug(f"Request Handlers : {request_handlers}")
        self.write_debug(f"Loaded {len(ies)} extractors")
        self.ping_and_speed()

    def _fmt_speed(self, bps: float) -> str:
        for unit in ("B/s", "KiB/s", "MiB/s", "GiB/s"):
            if bps < 1024:
                return f"{bps:.2f} {unit}"
            bps /= 1024
        return f"{bps:.2f} TiB/s"

    def estimate_speed(self):
        try:
            url = "https://speed.cloudflare.com/__down?bytes=5000000"
            t0 = time.perf_counter()
            r = self.extractor_base._request(url)
            size = len(r.content)
            dt = time.perf_counter() - t0
            bps = size / max(dt, 1e-6)
            speed_str = self._fmt_speed(bps)
            self.write_debug(f"Downloading Speed: {speed_str}")
        except Exception as e:
            self.write_debug(f"Downloading Speed test fail {e}")

    def ping_and_speed(self):
        if not self.verbose:
            return
        if self.verbose < 2:
            return

        url = "https://www.google.com/generate_204"
        try:
            t0 = time.perf_counter()
            self.extractor_base._request(url)
            ping_ms = (time.perf_counter() - t0) * 1000
            self.write_debug(f"Ping: {ping_ms:.1f} ms")
            self.estimate_speed()
        except Exception as e:
            self.write_debug(f"Ping fail {e}")

    def show_list_extractors(self):
        if self.list_extractors is False:
            return
        for extractor in self.extractor_base.gen_extractors(self):
            extractor = extractor._IE_NAME or extractor.__name__
            self.write(f"{extractor}")

        return 0

    def list_impersonate_targets(self):
        try:
            from .networking._curl_cffi import ImpersonateTargets

            if not self.list_impersonate:
                return
            ImpersonateTargets(self.logger).show_impersonate_table()
        except ImportError:
            raise NobiDLError(
                'Curl_cffi not installed please install curl_cffi - command "pip install curl_cffi"',
                expected=True,
            )
        return 0

    def request_handlers(self):
        request_handlers = set()
        try:
            import requests

            request_handlers.add(requests.__name__.capitalize())
        except Exception:
            pass

        try:
            import curl_cffi

            request_handlers.add(curl_cffi.__name__.capitalize())
        except Exception:
            pass

        try:
            import urllib

            request_handlers.add(urllib.__name__.capitalize())
        except Exception:
            pass

        return request_handlers

    def resolve_info_dict(self, info_dict_or_entry):
        formats = info_dict_or_entry.get("formats")
        resolved_formats = Resolve_FMTS(self, info_dict_or_entry, None, formats)()
        info_dict_or_entry.pop("formats")
        return {**info_dict_or_entry, "formats": resolved_formats}

    def handle_playlist(self, info_dict):
        resolved_entries = []
        entries = info_dict.get("entries", [])

        items = len(entries)
        title = info_dict.get("title")

        self.write(f"[info] Downloading Series: {title}")
        self.to_screen(f"Playlist {title} Downloading {items} items of {items}")

        for item, entry in enumerate(entries, start=1):
            self.to_screen(
                f"Downloading item {ascii_color(item, 32)} of {ascii_color(items, 34)}"
            )

            if self.opts.dump_json:
                resolved_entry = self.resolve_info_dict(entry)
                resolved_entries.append(resolved_entry)
                self.write(json.dumps(resolved_entry))
            if self.opts.dump_json:
                playlist = {
                    **info_dict,
                    "entries": resolved_entries,
                }
                self.write(json.dumps(playlist))

            elif self.opts.list_formats:
                formats = entry.get("formats")
                title = entry.get("title") or "Unkown title"
                if not formats:
                    pass
                self.write(f"[info] Downloading {title}")
                self.write(render_formats_table(formats, entry))
                continue

            self.downloader(self, entry, self.opts, self.logger)()

        return None

    def gen_info_dict(self, info):
        info.setdefault("fulltitle", info.get("title"))
        info.setdefault("thumbnails", [])
        info.setdefault("series", bool(info.get("entries")))
        info.setdefault("formats", [])
        info.setdefault("ie", self._ie)

        for f in info["formats"]:
            f.setdefault("protocol", "https")
            f.setdefault("vcodec", None)
            f.setdefault("acodec", None)
            f.setdefault("fps", None)
            f.setdefault("filesize", None)
            f.setdefault("filesize_approx", None)
            f.setdefault("abr", None)
            f.setdefault("tbr", None)
            f.setdefault("language", None)
            f.setdefault("audio_channels", None)
            headers = f.get("http_headers") or f.get("https_headers")
            if headers:
                f.setdefault("http_headers", headers)

        return info

    def _real_extract(self, url=None):
        if self.list_impersonate:
            return self.list_impersonate_targets()

        if self.list_extractors:
            return self.show_list_extractors()

        if not self.opts.search and not self.args:
            self.opt_error("No url or search query provided")
            return

        url = None
        q = None

        if self.opts.search:
            q = self.opts.search

        elif self.args:
            arg = self.args[0]
            if arg.startswith(("http://", "https://")):
                url = arg
            else:
                q = arg

        if q:
            result = self.searcher.search(q)
            best = result.get("best")
            if not best:
                self.to_screen(f"[info] No Result found for {q}")
                return
            url = best.get("url")

        if not url:
            self.to_screen("[error] No URL provided")
            return

        ie = ExtractorBase.get_extractor(self, url)
        if not ie:
            self.raise_error(f"Unsupported url: {url}")
            return

        self._ie = ie
        self.logger.set_ie(self._ie)
        self.to_screen(f"Extracting url: {url}")
        info = self._ie.extract(url)
        if not info or info is None:
            return self.show_warning(
                "Extractor return nothing there is an error in Extractor."
            )
        info_dict = self.gen_info_dict(info)
        if info_dict.get("_type") == "series":
            return self.handle_playlist(info)

        if not info_dict.get("formats"):
            raise NobiDLError("No Formats Found", url=url)

        if self.opts.list_formats and info_dict.get("formats"):
            title = info_dict.get("title")
            self.write(f"[info] Available formats for {title}")
            self.write(render_formats_table(info_dict["formats"], info_dict))
            pass

        if self.opts.dump_json:
            resolved_info = self.resolve_info_dict(info_dict)
            self.write(json.dumps(resolved_info))
            pass

        if not self.opts.list_formats and info_dict.get("formats"):
            self.downloader(self, info_dict, self.opts, self.logger)()
            return

        return info_dict


def main():
    opts, args, opt_error = options()
    md = NobiDL(opts, args, opt_error)
    ExtractorBase(md)
    try:
        md._real_initialize()
        md._real_extract()

    except KeyboardInterrupt:
        md.write("Interrupted by user")
        raise SystemExit(130)

    except NobiDLError as e:
        if opts.verbose:
            raise e
        md.write(e)



import os
import time

from ..networking import Request
from ..utils import filename_from_title
from .common import ProgressPrinter


class HttpDownloader:
    def __init__(self, format, info_dict, options, logger):
        self.format = format
        self.info_dict = info_dict
        self.formats = info_dict.get("formats") or []
        self.opts = options
        self.logger = logger
        self.stdout = logger.to_stdout
        self.continue_dl = self.opts.continue_dl

    def to_screen(self, msg):
        return self.stdout(f"[http_downloader] {msg}")

    def __call__(self):
        return self.downloader()

    def _prepare_download(self, fmt):
        title = self.info_dict.get("title")
        url = fmt["url"]
        ext = fmt.get("ext", "mp4")

        filename = filename_from_title(title, url) if title else f"Unknown movie.{ext}"
        part = filename + ".part"

        resume_from = 0
        if self.continue_dl and os.path.exists(part):
            resume_from = os.path.getsize(part) or 0

        return filename, part, resume_from

    def _open_response(self, url, resume_from, headers):
        if resume_from:
            headers["Range"] = f"bytes={resume_from}-"

        resp = Request(
            url=url,
            logger=self.logger,
            method="GET",
            stream=True,
            headers=headers,
        )()

        if resume_from and getattr(resp, "status_code", None) == 200:
            headers.pop("Range", None)
            resp = Request(
                url=url,
                logger=self.logger,
                method="GET",
                stream=True,
                headers=headers,
            )()

        return resp

    def _detect_total_size(self, resp, resume_from, fmt):
        hdrs = getattr(resp, "headers", {}) or {}

        cr = hdrs.get("Content-Range") or hdrs.get("content-range")
        if cr and "/" in cr:
            try:
                return int(cr.split("/")[-1])
            except ValueError:
                pass

        cl = hdrs.get("Content-Length") or hdrs.get("content-length")
        if cl:
            try:
                size = int(cl)
                return size + resume_from if resume_from else size
            except ValueError:
                pass

        for value in ("filesize", "filesize_approx"):
            value = fmt.get(value)
            if isinstance(value, (int, float)) and value > 0:
                return int(value)

        return None

    def _download_stream(self, resp, part, resume_from, total, filename):
        progress = ProgressPrinter(self.logger)
        downloaded = resume_from
        ema_speed = None
        t0 = last_t = time.time()
        last_b = downloaded
        alpha = 0.15

        iterator = (
            resp.iter_bytes(1024 * 256)
            if hasattr(resp, "iter_bytes")
            else resp.iter_content(chunk_size=1024 * 256)
        )

        with open(part, "ab" if resume_from else "wb") as f:
            for chunk in iterator:
                if not chunk:
                    continue

                f.write(chunk)
                downloaded += len(chunk)

                now = time.time()
                if now - last_t >= 0.2:
                    dt = now - last_t
                    db = downloaded - last_b
                    speed = db / max(dt, 1e-6)

                    ema_speed = (
                        speed
                        if ema_speed is None
                        else (alpha * speed + (1 - alpha) * ema_speed)
                    )

                    eta = (
                        (total - downloaded) / ema_speed
                        if total and ema_speed
                        else None
                    )

                    progress(
                        {
                            "status": "downloading",
                            "filename": filename,
                            "downloaded_bytes": downloaded,
                            "total_bytes": total,
                            "speed": ema_speed,
                            "eta": eta,
                            "elapsed": now - t0,
                            "progress_type": "bytes",
                        }
                    )

                    last_t = now
                    last_b = downloaded

        return downloaded, time.time() - t0

    def downloader(self):
        fmt = self.format
        filename, part, resume_from = self._prepare_download(fmt)

        self.to_screen(f"Downloading format: {fmt.get('format_id')}")
        self.to_screen(f'Invoking [http] Downloader on "{fmt["url"]}"')
        self.to_screen(f"Destination: {filename}")

        headers = dict(fmt.get("https_headers") or {})
        resp = None

        try:
            resp = self._open_response(fmt["url"], resume_from, headers)
            total = self._detect_total_size(resp, resume_from, fmt)

            downloaded, elapsed = self._download_stream(
                resp, part, resume_from, total, filename
            )

            os.replace(part, filename)

            ProgressPrinter(self.logger)(
                {
                    "status": "finished",
                    "filename": filename,
                    "downloaded_bytes": downloaded,
                    "total_bytes": total,
                    "elapsed": elapsed,
                    "progress_type": "bytes",
                }
            )

        except KeyboardInterrupt:
            self.to_screen(
                "Interrupted by user. add -c or --continue for continue downloading."
            )
            raise SystemExit(130)

        finally:
            if resp and hasattr(resp, "close"):
                resp.close()



import re

from ..extractors.hdhub import Hdhub4uME
from ..nobi_dl import ExtractorBase
from ..utils import (
    _get_qparam,
    _og_title,
    _parse_a_tag,
    _parse_a_tags,
    _parse_resolution,
    _search_regex,
    b64d,
    clean_url,
    determine_ext,
    determine_filesize,
    random_id,
    url_filename,
)


class Resolve_FMTS(ExtractorBase):
    def __init__(self, md, info_dict, fmt: str = None, formats: dict = None):
        super().__init__(md)
        self.formats = formats
        self.fmt = fmt
        self._ie = info_dict.get("ie", None) if info_dict else None

    _GLOBAL_SEEN_URLS = set()

    @property
    def hdhubme(self):
        return Hdhub4uME(self.md)

    def _google_resolver(self, url, need_dict=False):
        if "google" not in url:
            response = self._request(url, method="HEAD")
            url = _search_regex(
                r"""(?:href[^'"]+["'])?(https://?[^/]+googleuser[^"'?#]+)(?:['"])?""",
                response.url,
            )
            if "google" not in url:
                return {} if need_dict is True else None
        google_head = self._request(url, method="HEAD", note="Ping Google server")
        if google_head.status_code != 200:
            return {}
        headers = google_head.headers
        cd = headers.get("Content-Disposition", "")
        filename = _search_regex(r'filename\s*=\s*"([^"]+)"', cd)
        data = {
            "url": url,
            "ext": determine_ext(url, filename),
            "acodec": True,
            "vcodec": True,
            "format_id": random_id(),
            **_parse_resolution(filename),
            **determine_filesize(headers),
        }
        if need_dict is True:
            return [data]
        return data

    def fast_dl(self, url):
        fast_data = self._request(
            url, method="POST", note="Downloading Fast DL data"
        ).text
        google_url = _search_regex(
            r"""href[^'"]+["'](https://?[^/]+googleuser[^"']+)['"]""", fast_data
        )
        if not google_url:
            return None
        return self._google_resolver(google_url, True)

    def _parse_pixel(self, url, format_id=None):
        API = "{}/api/file/{}?download"
        if "pixel" not in url:
            return []
        self._GLOBAL_SEEN_URLS.add(url)

        if "?id=" in url:
            req = self._request(url, method="HEAD", note="ping pixel server")
            google_url = _get_qparam(req.url, ("url", "url"))
            if not google_url:
                return []
            return [self._google_resolver(google_url, need_dict=True)]
        elif "?id" not in url:
            webpage = self._download_webpage(
                url, note="Downloading pixel server information"
            )
            d_vid = re.search(
                r"(https?://)?(?P<domain>[^/]+)/.+/(?P<vid>[^/?&#]+)", url
            )
            if not d_vid:
                return []
            title = _og_title(webpage)
            domain, vid = d_vid.group("domain", "vid")
            host = f"https://{domain}" if "http" not in domain else domain
            url = API.format(host, vid)

            return [
                {
                    "url": url,
                    "format_id": format_id or title,
                    **_parse_resolution(title),
                    "acodec": True,
                    "vcodec": True,
                    "ext": determine_ext(url),
                    **determine_filesize(webpage),
                }
            ]

        return []

    def _parse_fsl(self, url: str, format_id: str | None = None):
        req = self._request(url, method="HEAD", note="ping FSL server")
        if getattr(req, "status_code", 0) != 200:
            return []
        real_url = getattr(req, "url", url) or url
        if "google" in real_url:
            return [self._google_resolver(real_url, True)]
        headers = req.headers
        cd = (
            headers.get("Content-Disposition")
            or headers.get("content-disposition")
            or ""
        )
        filename = None
        try:
            filename = _search_regex(
                r'filename\*\s*=\s*[^\'"]*\'\'([^;]+)|filename\s*=\s*"([^"]+)"', cd
            )
            if isinstance(filename, (list, tuple)):
                filename = next((x for x in filename if x), None)
        except Exception:
            filename = None
        filename = filename or url_filename(real_url) or url_filename(url) or "download"
        out_format_id = (format_id or filename or "download").strip()
        return [
            {
                "url": real_url,
                "format_id": out_format_id,
                **_parse_resolution(filename),
                "acodec": True,
                "vcodec": True,
                "ext": determine_ext(real_url),
                **determine_filesize(headers),
            }
        ]

    def _parse_direct_drive(self, url):
        formats = []
        domain = _search_regex(
            r'(https?://[^"\s<>]*drive[^"/\s<>]*)', url, default=None
        )
        if not domain:
            return []

        webpage = self._download_webpage(url, note="Downloading hubdrive webpage")

        did = None
        if not did:
            did = _search_regex(
                r'\bid\s*=\s*["\']down-id["\'][^>]*>\s*([^<\s]+)', webpage, default=None
            )
        if not did:
            did = _search_regex(r"\bdown-id[^>]*>\s*([^<\s]+)", webpage, default=None)

        if not did:
            return []

        data = self._download_json(
            f"{domain}/ajax.php?ajax=direct-download",
            data={"id": str(did)},
            headers={
                "Origin": domain,
                "Referer": f"{domain}/",
                "X-Requested-With": "XMLHttpRequest",
            },
            timeout=30,
        )

        if str(data.get("code")) != "200":
            self.to_screen("retrying for extracting hubdrive information")
            data = self._download_json(
                f"{domain}/ajax.php?ajax=direct-download",
                data={"id": str(did)},
                headers={
                    "Origin": domain,
                    "Referer": f"{domain}/",
                    "X-Requested-With": "XMLHttpRequest",
                },
                timeout=30,
            )

        if str(data.get("code")) != "200":
            return []

        d_data = data.get("data") or {}
        format_id = d_data.get("n") or ""

        gd_url = d_data.get("gd")
        file_path = data.get("file")
        if gd_url:
            formats.append(self._google_resolver(gd_url, True))

        if file_path:
            dl_url = (
                file_path if file_path.startswith("http") else f"{domain}{file_path}"
            )
            worker_webpage = self._download_webpage(
                dl_url, "Downloading hubdrive worker webpage"
            )
            direct_url = _search_regex(
                r'(https?://[^"\s<>]*workers\.dev/[^"\s<>]+)',
                worker_webpage,
                default=None,
            )
            if direct_url:
                h = self._request(direct_url, method="HEAD")
                real = getattr(h, "url", direct_url) or direct_url
                formats.append(
                    {
                        "format_id": format_id,
                        "url": real,
                        **_parse_resolution(format_id or worker_webpage),
                        "acodec": True,
                        "vcodec": True,
                        "ext": determine_ext(real),
                        **determine_filesize(h.headers),
                    }
                )

        return formats

    def _tech_unblock_decoder(self, url):
        if "tech" not in url:
            return None
        domain, value = re.search(
            r"https://?(?P<domain>[^/]+)[^\bid]+id\s*=\s*(?P<token>[^?&#]+)", url
        ).groups()
        domain = f"https://{domain}" if "://" not in domain else domain
        req = self._request(
            domain,
            data={"_wp_http": value},
            method="POST",
            note="Downloading tech unblock webpage",
        )
        url, name, post_value = re.search(
            r'''<form[^>]+action\s*=\s*['\"](?P<url>[^'\"]+)"?[^>]+>[^>]+name\s*=\s*['\"]+(?P<name>[^'\"]+)"[^>]+value\s*=\s*['\"]+(?P<value>[^'\"]+)"''',
            req.text,
        ).groups()
        req2 = self._request(
            url,
            data={name: post_value},
            method="POST",
            note="Redirecting from tech unblock",
        )
        url, cname, cvalue = re.search(
            r"""setAttribute.+\b(?P<url>https[^'\"]+)[\s\S]+?s[^'\"]+'(?P<cname>[^'\"]+)'[^'\"]+'(?P<cvalue>[^'\"]+)""",
            req2.text,
        ).groups()
        req3 = self._request(
            url,
            cookies={cname: cvalue},
            method="POST",
            note="Downloading Final webpage of tech unblock",
        )
        final_url = re.search(r'''url\s*=\s*(?:['\"]+)?([^'\"]+)?"''', req3.text).group(
            1
        )
        return final_url if final_url else None

    def modpro(self, url):
        webpage = self._download_webpage(url)
        for atag in _parse_a_tags(webpage):
            href, name = _parse_a_tag(atag)
            label = (name or "").lower()
            if not any(
                k in label
                for k in ("fast server", "st serve", "google drive", "gle drive")
            ):
                continue
            if "fast" in label:
                return self._tech_unblock_decoder(href) if "tech" in href else href
            elif "lr drive" in label:
                return self._tech_unblock_decoder(href) if "tech" in href else href
            else:
                return href
        return None

    def _parse_hubcloud(self, url):
        if not any(k in url for k in ("hub", "cloud", "id=")):
            return []
        if "?id" in url:
            url = self.hdhubme._hdhub_url_decoder(url)
        self._GLOBAL_SEEN_URLS.add(url)
        webpage = self._download_webpage(url, "Downloading hubcloud information")
        cdns_url = _search_regex(
            r'<a[^>]+id\s*=\s*"download[^>]+href\s*=\s*"([^"]+)"[^>]*>[^<]*',
            webpage,
            default=None,
        )
        if not cdns_url:
            return []

        cdn_webpage = self._download_webpage(cdns_url)
        formats = []

        for a_tag in _parse_a_tags(cdn_webpage):
            href, _ = _parse_a_tag(a_tag)
            if not href:
                continue

            check = href.lower()
            if not any(k in check for k in ("pixel", "hubcdn", "fsl")):
                continue

            if "pixel" in check:
                return self._parse_pixel(href)
            elif "fsl" in check:
                return self._parse_fsl(href)
        return formats

    def drive_seed(self, url):
        if "modpro" in url or "archives" in url:
            url = self.modpro(url)
        elif "tech" in url:
            url = self._tech_unblock_decoder(url)
        elif "driveseed" in url:
            url = url
        domain = clean_url(url)
        webpage = self._download_webpage(url, note="Downloading drive seed webpage")
        if "/file" not in url:
            path = _search_regex(r"""['\"]+((?:https://)?/?fil[^'\"]+)""", webpage)
            url = f"{domain}{path}" if "http" not in path else path
            webpage = self._download_webpage(
                url, note="Downloading drive seed Redirected webpage"
            )

        for atag in _parse_a_tags(webpage):
            href, name = _parse_a_tag(atag)
            label = (name or "").lower()
            if not any(k in label for k in ("instant download", "direct links")):
                continue
            if "instant" in label:
                return self._google_resolver(href)
            elif "driveseed" in href:
                return self.driveseed_direct(href)
            else:
                return href

    def bollyflix_formats(self, url, format_id=None):
        qualties = []
        url = self.bolly_url_decoder(url) if "?id=" in url else url
        webpage = self._download_webpage(url)
        for atag in _parse_a_tags(webpage):
            href, name = _parse_a_tag(atag)
            format_id = (name or "").lower()
            if not any(
                k in format_id
                for k in ("instant", "fast cloud", "zipdisk", "pixel", "gofile")
            ):
                continue
            if "pixel" in format_id:
                self._parse_pixel(href)
            elif "instant" in format_id:
                return self._google_resolver(href)
            elif "zipdisk" in format_id:
                continue

        return qualties

    def bolly_url_decoder_resolver(self, url):
        webpage = self._download_webpage(
            url, note="Downloading Bolly URL Decoder Webpage"
        )
        out = []
        for p1, p2 in re.findall(
            r"""(?:'(?P<p1>[^']*)'\s*\+\s*)+'(?P<p2>[^']*)'?""", webpage
        ):
            parts = p1 + p2
            if "tech-news" in parts:
                out.append("https://box.tech-news.app")
            elif ".com" in parts:
                parts = f"https://{parts}" if "https" not in parts else parts
                out.append(parts)
        if not out:
            return None
        for domain in out:
            try:
                if self._ping_host(domain) == 200:
                    return domain
            except Exception:
                pass
        return out[0]

    def bolly_url_decoder(self, url):
        domain = self.bolly_url_decoder_resolver(url)
        if domain is None:
            webpage = self._download_webpage(
                url, note="Downloading Bolly URL Decoder Webpage"
            )
        token = _search_regex(r"\bid\s*=([^/?&#]+)", url)
        webpage = self._download_webpage(
            f"{domain}/?id={token}", headers={"referer": clean_url(url)}
        )
        e_token = _search_regex(r'\blink\s*?"\s*:\s*"([^"]+)', webpage)
        if not e_token:
            return None
        return b64d(e_token)

    def driveseed_direct(self, url):
        formats = []
        if "drive" not in url:
            return None
        url = f"{url}?type=1" if "type=1" not in url else url
        webpage = self._download_webpage(url).text
        if "available for this file" in webpage.lower():
            webpage = self._download_webpage(url.replace("type=1", "type=2"))
        for atag in _parse_a_tags(webpage):
            href, name = _parse_a_tag(atag)
            label = (name or "").lower()
            if "download" not in label:
                continue
            if "dev" not in href:
                continue
            formats.append(
                {
                    "format_id": random_id(),
                    "url": href,
                    "acodec": True,
                    "vcodec": True,
                    "ext": determine_ext(href, href),
                    **determine_filesize(webpage),
                    **_parse_resolution(webpage),
                }
            )

    def nextdrive(self, url):
        webpage = self._download_webpage(url, "Downloading NextGen Webpage")
        for atag in _parse_a_tags(webpage):
            href, _ = _parse_a_tag(atag)
            if not href:
                continue
            if not any(k in href for k in ("fast-dl", "fast-d", "st-dl")):
                continue
            return self.fast_dl(href)

        return []

    def resolve_provider(self, url):
        if "pixel" in url:
            return self.pixel_cdn(url)
        elif "fsl" in url:
            return self._parse_fsl(url)
        elif any(k in url for k in ("nextdrive", "extdrive", "exdrive")):
            return self.nextdrive(url)
        elif any(k in url for k in ("hubcloud", "web", "gadgetsweb")):
            return self._parse_hubcloud(url)
        elif any(k in url for k in ("driveseed", "tech", "modpro")):
            return self.drive_seed(url)
        elif "hubdrive" in url:
            return self._parse_direct_drive(url)
        elif "fast" in url:
            return self.fast_dl(url)
        elif any(k in url for k in ("?id=", "gdflix")):
            return self.bollyflix_formats(url)
        return url

    def resolve_format(self, fmt: dict) -> list[dict]:
        url = fmt.get("url")
        if not url:
            return []

        if "m3u8" in url:
            return [fmt]

        fmts = self.resolve_provider(url)

        if not fmts:
            return []

        if isinstance(fmts, list):
            return fmts

        if isinstance(fmts, dict):
            return [fmts]

        return [fmt]

    def __call__(self):
        return self._real_solver()

    def _real_solver(self) -> list[dict]:
        self.write_debug("Lazy formats Resolving")
        if self._ie:
            self.logger.set_ie(self._ie, "lazy resolver")
        if self.formats:
            resolved_formats = []
            for fmt in self.formats:
                fmt = self.resolve_format(fmt)
                resolved_formats.extend(fmt)
            return resolved_formats

        normal_dict_fmt = self.resolve_format(self.fmt)
        return (
            normal_dict_fmt[0]
            if isinstance(normal_dict_fmt, list) and self.fmt
            else normal_dict_fmt
        )



import itertools
import re
from urllib.parse import quote_plus

from nobi_dl import ExtractorBase
from nobi_dl.utils import (
    _og_thumbnail,
    _og_title,
    _parse_a_tag,
    _parse_a_tags,
    _parse_resolution,
    _search_regex,
    determine_ext,
    determine_filesize,
    fix_entries,
    is_series,
    random_id,
)

from ..downloader.fmt_resolver import Resolve_FMTS
from .hdhub import Hdhub4uME


class BollyFlixME(ExtractorBase):
    _IE_NAME = "Bollyflix"
    _VALID_URL = r"^https?://.*bollyflix\..+"
    _SEARCH = True

    @property
    def HOST(self):
        return self.host_finder.bollyflix()

    @property
    def hdhub4u(self):
        return Hdhub4uME(self.md)

    @property
    def bolly_url_decoder(self):
        return Resolve_FMTS(self.md, None, None, None).bolly_url_decoder

    def _parse_content(self, html):
        results = []
        for article in re.findall(r"<article[\s\S]+?<\/article>", html):
            href = _search_regex(r'href="([^"]+)"', article)
            href = f"{self.HOST}{href}" if "https" not in href else href
            thumbnail = _search_regex(r'src="([^"]+)"', article)
            thumbnail = (
                f"{self.HOST}{thumbnail}" if "https" not in thumbnail else thumbnail
            )
            title = _search_regex(r'title="([^"]+)"', article)
            results.append(
                {
                    "title": title,
                    "thumbnail": thumbnail,
                    "url": href,
                    "acodec": True,
                    "vcodec": False,
                }
            )
        return results

    def _real_search(self, query):
        results = []
        for page in itertools.count(1):
            page = self._download_webpage(
                f"{self.HOST}/search/{quote_plus(query)}/page/{page}",
                headers={
                    "referer": "https://bollyflix.to/",
                },
            )
            results.extend(self._parse_content(page))
            next_page = _search_regex(r"<a[^>]+next[^>]+>Next<\/a>", page)
            if not next_page:
                break
        return results

    def get_movie(self, url_or_webpage):
        formats = []
        if "<html" in url_or_webpage:
            webpage = url_or_webpage
        else:
            webpage = self._download_webpage(url_or_webpage)
        for h5 in re.findall(r"<h5[\s\S]+?<\/h5>[\s\S]+?<a[\s\S]+?<\/a>", webpage):
            if "<a" not in h5:
                continue
            for atag in _parse_a_tags(h5):
                href, name = _parse_a_tag(atag)
                label = (name or "").lower()
                if not any(
                    k in label for k in ("google drive", "le drive", "ogle drive")
                ):
                    continue
                formats.append(
                    {
                        "format_id": random_id(),
                        "url": href,
                        "acodec": True,
                        "vcodec": True,
                        "ext": determine_ext(href, href),
                        **determine_filesize(h5),
                        **_parse_resolution(h5),
                    }
                )

        return {
            "title": _og_title(webpage),
            "thumbnail": _og_thumbnail(webpage),
            "formats": formats,
        }

    def block_parser(self, webpage, h4, url):
        entries = []

        fmt_url = self.bolly_url_decoder(url) if "?id=" in url else url
        fmt_webpage = self._download_webpage(fmt_url)
        for atag in _parse_a_tags(fmt_webpage):
            href, name = _parse_a_tag(atag)
            if not name:
                continue
            if not any(k in name.lower() for k in ("episode", "epis")):
                continue
            ep_no = _search_regex(
                r"(?is)(?:episode|ep0|ep)\s*(\d+)", name, default=None
            )
            season = _search_regex(r"(?is)(?:season|s|s0)\s*(\d+)", h4, default=None)
            entries.append(
                {
                    "title": f"{_og_title(webpage)} - Season {season} - Episode {ep_no}",
                    "season": season,
                    "episode": ep_no,
                    "formats": [
                        {
                            "url": href,
                            "format_id": random_id(),
                            "acodec": True,
                            "vcodec": False,
                            "ext": determine_ext(href, href),
                            **determine_filesize(h4),
                            **_parse_resolution(h4),
                        }
                    ],
                }
            )

        return entries

    def series_formats(self, webpage):
        entries = []
        for h4 in re.findall(r"<h4[\s\S]+?<\/h4>\s*<p[\s\S]+?<\/p>", webpage):
            if "<a" not in h4:
                continue
            for atag in _parse_a_tags(h4):
                href, name = _parse_a_tag(atag)
                if not any(
                    k in name.lower()
                    for k in (
                        "download links",
                        "download link",
                        "load links",
                        "load link",
                    )
                ):
                    continue
                if href is None:
                    continue
                entries.extend(self.block_parser(webpage, h4, href))

        return entries

    def _get_series(self, url_or_webpage, url):
        if "<html" in url_or_webpage:
            webpage = url_or_webpage
        else:
            webpage = self._download_webpage(url_or_webpage)

        entries = fix_entries(self, self.series_formats(webpage))

        return self.playlist_result(
            entries=entries,
            url=url,
            playlist_title=_og_title(webpage),
        )

    def _real_extract(self, url):
        webpage = self._download_webpage(url)
        if is_series(url, webpage):
            return self._get_series(webpage, url)
        else:
            return self.get_movie(webpage)



import optparse
import os
import sys


def options():
    prog = os.path.basename(sys.argv[0])

    p = optparse.OptionParser(
        usage=f"{prog} [options] [url]",
        description="nobi-dl – scrape-based movie & series downloader",
    )

    p.add_option(
        "--search",
        dest="search",
        metavar="QUERY",
        help="Search for a movie or series instead of providing a direct URL",
    )

    p.add_option(
        "-v",
        "--verbose",
        action="count",
        dest="verbose",
        default=0,
        help="Increase verbosity (-v, -vv)",
    )

    p.add_option(
        "-q",
        "--quiet",
        action="store_true",
        dest="quiet",
        default=False,
        help="Suppress non-essential output",
    )

    p.add_option(
        "--parse-network",
        action="store_true",
        dest="parse_network",
        default=False,
        help="Parse network requests to extract hidden media URLs",
    )

    p.add_option(
        "-F",
        "--list-formats",
        action="store_true",
        dest="list_formats",
        default=False,
        help="List all available formats without downloading",
    )

    p.add_option(
        "--impersonate",
        dest="impersonate",
        metavar="BROWSER",
        default=None,
        help="Impersonate a browser (e.g. chrome, firefox, safari)",
    )

    p.add_option(
        "--print-traffic",
        action="store_true",
        dest="print_traffic",
        default=False,
        help="Print sent and received HTTP traffic",
    )

    p.add_option(
        "-j",
        "--dump-json",
        action="store_true",
        dest="dump_json",
        default=False,
        help="Dump extracted info as JSON to stdout",
    )

    p.add_option(
        "--list-impersonate-targets",
        action="store_true",
        dest="list_impersonate",
        default=False,
        help="List all supported impersonation targets",
    )

    p.add_option(
        "-c",
        "--continue",
        action="store_true",
        dest="continue_dl",
        default=False,
        help="Resume partially downloaded files if possible",
    )

    p.add_option(
        "--list-extractors",
        action="store_true",
        dest="list_extractors",
        default=False,
        help="List all available extractors",
    )

    p.add_option(
        "--extractor",
        dest="force_extractor",
        metavar="NAME",
        default=None,
        help="Force a specific extractor to be used",
    )

    p.add_option(
        "--version",
        action="store_true",
        dest="show_version",
        help="Show program version and exit",
    )

    opts, args = p.parse_args()

    if opts.show_version:
        from nobi_dl.version import __version__

        p.exit(0, f"nobi-dl {__version__}\n")

    opts.movie_url = args[0] if (args and not opts.search) else None

    if (
        not opts.movie_url
        and not opts.search
        and not opts.list_impersonate
        and not opts.list_extractors
    ):
        p.error("Provide a URL or use --search QUERY")

    return opts, args, p.error



import ast
import itertools
import re
from html import unescape

from nobi_dl.common import ExtractorBase
from nobi_dl.utils import (
    _og_thumbnail,
    _og_title,
    _parse_a_tag,
    _parse_a_tags,
    _parse_resolution,
    _remove_duplicate,
    _search_json,
    _search_regex,
    b64d,
    determine_ext,
    determine_filesize,
    fix_entries,
    is_series,
    random_id,
    shifter,
)

from ..host_finder import HOST_FINDER


class Hdhub4uME(ExtractorBase):
    _IE_NAME = "hdhub4u"
    _VALID_URL = r"https://?[^\.]+\.hdhub4u[^/]+/[^?&#]+"
    _SEARCH = True

    _GLOBAL_SEEN_URLS = set()
    _VALID_DRIVE_URL = (
        r'(https?://[^"\s<>]*hubdrive[^"\s<>]*/(?:file/)?(?:\d+|[^"\s<>]+))'
    )
    _VALID_PIXEL_URL = r'(https?://[^"\s<>]*pixel[^"\s<>]*/(?:\d+|[^"\s<>]+))'
    _VALID_ARCHIVE_URL = r'(https?://[^"\s<>]*archives[^"\s<>]*/(?:\d+|[^"\s<>]+))'
    _VALID_HUBCLOUD_URL = (
        r'(https?://[^"\s<>]*cloud[^/\s<>]+/drive/(?!admin(?:/|$))[^"\s<>]+)'
    )

    @property
    def _search_host(self):
        return HOST_FINDER(self.md, self)._hdhub_search_host()

    def _real_search(self, query, limit=50):
        results = []

        for page in itertools.count(1):
            resp = self._download_json(
                f"{self._search_host}/collections/post/documents/search",
                query={
                    "q": query,
                    "query_by": "post_title,category,stars,director,imdb_id",
                    "limit": 15,
                    "use_cache": "true",
                    "page": page,
                    "highlight_fields": None,
                    "analytics_tag": "2026-01-17",
                    "query_by_weights": "4,2,2,2,4",
                },
                headers={
                    "referer": "https://hdhub4u.gd/",
                    "origin": "https://hdhub4u.gd",
                },
            )

            hits = resp.get("hits") or []
            for movie in hits:
                data = movie.get("document") or movie

                title = data.get("post_title")
                if not title:
                    continue

                url = data.get("permaurl")
                if not url:
                    continue

                if "https" not in url:
                    url = f"{self._search_host}{url}"

                results.append(
                    {
                        "title": title,
                        "url": url,
                        "thumbnail": data.get("post_thumbnail"),
                        "year": data.get("post_date"),
                        "quality": _parse_resolution(title),
                        "source": self._IE_NAME,
                    }
                )

            if resp.get("found") == len(results) or limit >= len(results):
                break

        return _remove_duplicate(results)

    def _hdhub_url_decoder(self, url):
        webpage = self._download_webpage(url, "Downloading challange webpage")
        o_token = _search_regex(
            r"""s\s*\(\s*['"]o['"]\s*,\s*['"]([^'"]+)['"]""", webpage, default=None
        )
        if not o_token:
            return None
        data = shifter(o_token)
        if not data:
            return None
        final_url = b64d(data["o"])
        return final_url

    def _minochinos_decoder_util(self, p: str, base: int, k: list[str]) -> str:
        DIGITS = "0123456789abcdefghijklmnopqrstuvwxyz"

        def base_n_to_int(s: str) -> int:
            n = 0
            for ch in s:
                n = n * base + DIGITS.index(ch)
            return n

        def repl(m: re.Match) -> str:
            tok = m.group(0)
            try:
                idx = base_n_to_int(tok)
            except ValueError:
                return tok
            if 0 <= idx < len(k) and k[idx]:
                return k[idx]
            return tok

        return re.sub(r"\b[0-9a-z]+\b", repl, p)

    def _parse_hubcloud(self, url):
        if (
            not _search_regex(self._VALID_HUBCLOUD_URL, url, default=None)
            or url in self._GLOBAL_SEEN_URLS
        ):
            return []
        self._GLOBAL_SEEN_URLS.add(url)
        webpage = self._download_webpage(url, "Downloading hubcloud information")
        cdns_url = _search_regex(
            r'<a[^>]+id\s*=\s*"download[^>]+href\s*=\s*"([^"]+)"[^>]*>[^<]*',
            webpage,
            default=None,
        )
        if not cdns_url:
            return []

        cdn_webpage = self._download_webpage(cdns_url)
        formats = []

        for a_tag in _parse_a_tags(cdn_webpage):
            href, _ = _parse_a_tag(a_tag)
            if not href:
                continue

            check = href.lower()
            if not any(k in check for k in ("pixel", "hubcdn", "fsl")):
                continue

            if _search_regex(self._VALID_PIXEL_URL, href, default=None):
                formats.append(
                    {
                        "format_id": random_id(),
                        "url": href,
                        "ext": determine_ext(href),
                        "acodec": True,
                        "vcodec": True,
                        **determine_filesize(webpage),
                        **_parse_resolution(webpage),
                    }
                )
            elif "fsl" in check:
                formats.append(
                    {
                        "format_id": random_id(),
                        "url": href,
                        "ext": determine_ext(href),
                        "acodec": True,
                        "vcodec": True,
                        **determine_filesize(webpage),
                        **_parse_resolution(webpage),
                    }
                )
        return formats

    def _parse_drive(self, url):
        if (
            not _search_regex(self._VALID_DRIVE_URL, url, default=None)
            or url in self._GLOBAL_SEEN_URLS
        ):
            return []
        self._GLOBAL_SEEN_URLS.add(url)
        webpage = self._download_webpage(url, "Downloading hubdrive information")
        hubcloud_url = _search_regex(self._VALID_HUBCLOUD_URL, webpage, default=None)
        phc = self._parse_hubcloud(hubcloud_url)
        return phc or []

    def _parse_archives(self, url):
        formats = []
        if (
            not _search_regex(self._VALID_ARCHIVE_URL, url, default=None)
            or url in self._GLOBAL_SEEN_URLS
        ):
            return []
        self._GLOBAL_SEEN_URLS.add(url)
        webpage = self._download_webpage(url, "Downloading hub archieve information")
        hub_url = _search_regex(self._VALID_HUBCLOUD_URL, webpage, default=None)
        if hub_url:
            hc = self._parse_hubcloud(hub_url)
        if hc:
            formats.extend(hc)
        return formats

    def _parse_formats(self, url):
        raw_formats = []
        if "drive.space" in url:
            raw_formats.extend(self._parse_drive(url))
        elif "/archives/" in url:
            raw_formats.extend(self._parse_archives(url))
        elif "cloud" in url:
            raw_formats.extend(self._parse_hubcloud(url))
        elif "pixel" in url:
            raw_formats.extend(self._parse_pixel(url))
        elif "hublinks" in url:
            raw_formats.extend(self.hublinks(url))
        if not raw_formats:
            return []
        formats = []
        seen_url = set()
        for quality in raw_formats:
            url = quality.get("url")
            if url in seen_url:
                continue
            seen_url.add(url)
            formats.append(quality)

        return formats

    def _minochinos_js_decoder(self, webpage: str):
        scripts = re.findall(
            r"<script\b[^>]*>(.*?)</script>", webpage, flags=re.DOTALL | re.I
        )
        js_text = next(
            (s for s in scripts if "eval(function(p,a,c,k,e,d)" in s.replace(" ", "")),
            None,
        )
        if not js_text:
            return None
        m = re.search(
            r"""eval\s*\(\s*function\s*\(\s*p\s*,\s*a\s*,\s*c\s*,\s*k\s*,\s*e\s*,\s*d\s*\)\s*\{.*?\}
                \s*\(\s*
                (?P<q1>['"])(?P<payload>(?:\\.|(?!\1).)*)(?P=q1)\s*,\s*
                (?P<base>\d+)\s*,\s*(?P<count>\d+)\s*,\s*
                (?P<q2>['"])(?P<dict>(?:\\.|(?!\4).)*)(?P=q2)\s*\.split\(\s*(?P<q3>['"])\|(?P=q3)\s*\)
                \s*\)\s*\)
            """,
            js_text,
            flags=re.DOTALL | re.VERBOSE | re.I,
        )
        if not m:
            return None

        p = ast.literal_eval(m.group("q1") + m.group("payload") + m.group("q1"))
        base = int(m.group("base"))
        k_raw = ast.literal_eval(m.group("q2") + m.group("dict") + m.group("q2"))
        k = k_raw.split("|")
        return self._minochinos_decoder_util(p, base, k)

    def _extract_player_formats(self, direct_url_or_webpage):
        formats = []
        if "stream" in direct_url_or_webpage and direct_url_or_webpage.startswith(
            "http"
        ):
            video_url = direct_url_or_webpage
        else:
            video_url = _search_regex(
                r'<a\b[^>]*href\s*=\s*["\'](https?://[^"\']*stream[^"\']*)["\']',
                direct_url_or_webpage,
            )
            if not video_url:
                return []
        video_webpage = self._download_webpage(video_url, "Downloading player url")
        title = _og_title(video_webpage)
        js_data = self._minochinos_js_decoder(video_webpage)
        if not js_data:
            self.show_warning("Unable to solve minochinos data")
            return []
        video_urls = _search_json(r"(?:urls|Urls)\s*=\s*(\{[\s\S]+?);", js_data)
        video_domain = _search_regex(
            r""">\s*[\s\S]+?(https.+)/(?:file|embed|download)/[^<"']""", video_webpage
        )
        video_domain = (
            f"http://{video_domain}" if "http" not in video_domain else video_domain
        )
        if not video_urls:
            return []
        for name, value in video_urls.items():
            if name.startswith("hls") and ".m3u8" in value:
                url = f"{video_domain}{value}" if "https" not in value else value
                formats.extend(self.extract_m3u8_formats(url, title))

        return formats

    def _get_movie(self, url_or_webpage):
        formats = []
        if "<html" in url_or_webpage:
            webpage = url_or_webpage
        else:
            webpage = self._download_webpage(url_or_webpage)

        languages = _search_regex(r'\bLanguage[^"]+?[^>]+>(.[^<]+)<', webpage)
        title = _search_regex(r'\bheadline\s*"\s*:\s*"([^"]+)"', webpage, default="")
        for a_tag in _parse_a_tags(webpage):
            href, text = _parse_a_tag(a_tag)
            valid_url = _search_regex(
                r'https?://(?!www\.googletagmanager\.com)[^"\s]+(?:\?id=[^"\s]+|/file/[^"\s]*)',
                href,
                default=None,
            )
            t = (text or "").strip().lower()
            if (
                t in ("watch online", "watch", "online", "hdhub4u")
                or href is None
                or not valid_url
            ):
                continue
            url = self._hdhub_url_decoder(href) if "id=" in href else href
            formats.extend(self._parse_formats(url))
        formats.extend(self._extract_player_formats(webpage))
        return {
            "title": title,
            "languages": languages,
            "thumbnail": _og_thumbnail(webpage),
            "formats": formats,
        }

    def hublinks(self, url):
        formats = []
        webpage = self._download_webpage(url, note="Download hublinks information")
        for h3_or_h5 in re.findall(r"<(?:h3|h5)[\s\S]+?<\/(?:h3|h5>)", webpage):
            h3_or_h5_title = _search_regex(r"<(?:h3|h5)[^>]+>([^<]+)<", h3_or_h5)
            if h3_or_h5_title:
                h3_or_h5_title = unescape(h3_or_h5_title.strip())
            if not h3_or_h5_title:
                continue
            for atag in _parse_a_tags(h3_or_h5):
                href, name = _parse_a_tag(atag)
                if any(k in href.lower() for k in ("hubcdn", "hubdrive")):
                    continue
                size = determine_filesize(name)
                if "hubcloud" in name.lower():
                    formats.extend(self._parse_hubcloud(href))
                elif size or "hubdrive" in href:
                    formats.append(
                        {
                            "format_id": random_id(),
                            "url": href,
                            "ext": "mkv",
                            "acodec": True,
                            "vcodec": True,
                            **size,
                            **_parse_resolution(name),
                        }
                    )

        return formats

    def diff_entries(self, webpage):
        valid_url_keys = ("drive", "hubdrive", "hdstream", "stream", "?id=", "hubcloud")
        entries = []
        title = _og_title(webpage)

        for atag in _parse_a_tags(webpage):
            href, name = _parse_a_tag(atag)
            if not href or not any(k in href for k in valid_url_keys):
                continue
            if name and "4k" in name.lower():
                continue
            href = self._hdhub_url_decoder(href) if "?id=" in href else href
            episode_formats = []
            episode_formats.extend(self._parse_formats(href))

            ep = _search_regex(
                r"(?is)(?:s\d+\s*)?(?:episode|ep|e)\s*(\d+)", name, default="1"
            )

            entries.append(
                {
                    "title": f"{title} - ep {ep}",
                    "season": 1,
                    "episode": int(ep),
                    "formats": episode_formats,
                }
            )

        return entries

    def entries(self, webpage):
        entries = []
        title = _og_title(webpage)

        blocks = re.split(r"<\s*hr\s*/?>", webpage, flags=re.I)

        for block in blocks:
            ep = _search_regex(r"(?i)episode\s*(\d+)", block, default=None)
            season = _search_regex(r"(?is)(?:season|s0|s)\s*(\d+)", block, default=1)
            if not ep:
                continue

            episode_formats = []

            for atag in _parse_a_tags(block):
                href, name = _parse_a_tag(atag)
                if not href or not name:
                    continue

                lname = name.lower()
                if lname not in ("drive"):
                    continue

                href = self._hdhub_url_decoder(href) if "?id=" in href else href

                episode_formats.extend(self._parse_formats(href))

            entries.append(
                {
                    "title": f"{title} - ep {ep}",
                    "season": season,
                    "episode": int(ep),
                    "formats": episode_formats,
                }
            )

        if not any(entry.get("formats") for entry in entries):
            return self.diff_entries(webpage)

        return entries

    def _get_series(self, url_or_webpage, url):
        if "<html" in url_or_webpage:
            webpage = url_or_webpage
        else:
            webpage = self._download_webpage(url_or_webpage)

        entries = fix_entries(self, self.entries(webpage))

        return self.playlist_result(
            entries,
            url,
            _og_title(webpage),
        )

    def _real_extract(self, url):
        webpage = self._download_webpage(url)

        if is_series(url, webpage):
            return self._get_series(webpage, url)
        else:
            return self._get_movie(webpage)



import sys
import time

from ..utils import (
    NobiDLError,
)


class BestFMT:
    def __init__(self, formats: list[dict]):
        self.formats = formats or []

    def __call__(self) -> dict:
        return self.best()

    def _score(self, f: dict) -> tuple:
        vcodec = f.get("vcodec")
        acodec = f.get("acodec")
        has_av = int(
            bool(vcodec and vcodec != "none") and bool(acodec and acodec != "none")
        )

        height = f.get("height") or 0
        tbr = f.get("tbr") or 0
        filesize = f.get("filesize") or 0

        return (has_av, height, tbr, filesize)

    def best(self) -> dict:
        if not self.formats:
            raise NobiDLError("No formats found")

        return max(self.formats, key=self._score)


class ProgressPrinter:
    _ANSI = {
        "reset": "\033[0m",
        "bold": "\033[1m",
        "dim": "\033[2m",
        "blue": "\033[94m",
        "yellow": "\033[93m",
        "green": "\033[92m",
        "red": "\033[91m",
        "white": "\033[97m",
    }

    def __init__(self, logger, *, is_tty=None):
        self.log = logger
        self.is_tty = sys.stderr.isatty() if is_tty is None else is_tty
        self._last = 0.0

    def _c(self, text: str, *styles: str) -> str:
        if not self.is_tty:
            return text
        codes = "".join(self._ANSI[s] for s in styles if s in self._ANSI)
        return f"{codes}{text}{self._ANSI['reset']}"

    def _fmt_bytes(self, n: int) -> str:
        units = ["B", "KiB", "MiB", "GiB", "TiB"]
        f = float(n)
        for u in units:
            if f < 1024 or u == units[-1]:
                return f"{f:.2f}{u}" if u != "B" else f"{int(f)}B"
            f /= 1024

    def _fmt_time(self, s: float | None) -> str:
        if s is None:
            return "--:--"
        s = int(s)
        if s < 0:
            return "--:--"

        d, s = divmod(s, 86400)
        h, s = divmod(s, 3600)
        m, s = divmod(s, 60)

        if d:
            return f"{d}d {h:02d}:{m:02d}:{s:02d}"
        if h:
            return f"{h:02d}:{m:02d}:{s:02d}"
        return f"{m:02d}:{s:02d}"

    def _fmt_count(self, cur: int, total: int | None) -> str:
        if total:
            return f"{cur}/{total}"
        return f"{cur}/???"

    def _fmt_speed_frag(self, sps: float | None) -> str:
        if not sps:
            return "?? seg/s"
        return f"{sps:.2f} seg/s"

    def __call__(self, p: dict):
        now = time.time()
        if p.get("status") == "downloading" and now - self._last < 0.2:
            return
        self._last = now

        ptype = p.get("progress_type", "bytes")

        tag = "[download]"

        if ptype == "fragments":
            done = int(p.get("downloaded_fragments") or 0)
            total = p.get("total_fragments")
            speed = p.get("speed")
            eta = p.get("eta")

            pct_raw = f"{(done * 100 / total):6.2f}%" if total else "   --.-%"
            count_s = self._fmt_count(done, total)
            speed_s = self._fmt_speed_frag(speed)
            eta_s = self._fmt_time(eta)

            line = (
                f"{tag} "
                f"{self._c(f'{pct_raw}', 'blue')} "
                f"{self._c(count_s, 'white')} "
                f"at {self._c(speed_s, 'green')} "
                f"ETA {self._c(eta_s, 'yellow')}"
            )

        else:
            done = int(p.get("downloaded_bytes") or 0)
            total = p.get("total_bytes")
            speed = p.get("speed")
            eta = p.get("eta")

            pct_raw = f"{(done * 100 / total):6.2f}%" if total else "   --.-%"
            done_s = self._fmt_bytes(done)
            total_s = self._fmt_bytes(total) if total else "???"
            speed_s = (self._fmt_bytes(int(speed)) + "/s") if speed else "??/s"
            eta_s = self._fmt_time(eta)

            line = (
                f"{tag} "
                f"{self._c(f'[{pct_raw}]', 'blue')} "
                f"{self._c(f'{done_s}/{total_s}', 'white')} "
                f"at {self._c(speed_s, 'green')} "
                f"ETA {self._c(eta_s, 'yellow')}"
            )

        if self.is_tty:
            self.log.to_stderr("\r" + line + "\033[K", only_once=False, newline=False)
        else:
            self.log.to_stderr(line, only_once=False)

        if p.get("status") in ("finished", "error") and self.is_tty:
            self.log.to_stderr("\n", only_once=False, newline=False)
